"""
An electrode manager that deals with the relationships between electrode
numbers and electrode positions.
It can be used while importing data, but also to manage electrode positions for
data that does not come with actual electrode positions.

## Possible future features

## Implemented features
    * [done] Generate electrode numbers from a set of electrode coordinates
    * [done] Add new positions, including a .assume_regular_electrodes
      (spacing, axis=(x|y|z)) function that conforms electrode numbers to a
      fixed distance of electrodes.
    * [done] Assume reciprocal measurement by reversing
    * [done] The electrode numbering should follow an intuitive heuristic, i.e.
      from left to right, top to bottom (think of rhizotrons or borehole
      settings). This behavior should be changeable by means of a custom sort
      function that takes (x,z) or (x,y,z) tuples and returns a given order.
    * [done] 2D and 3D
    * [done] .from_existing_ordering(iterable(numbers), iterable(coordinates))
    * [done] reverse lookup: given a fixed electrode assignment table, return
    * [done] test/account for positions not registered yet, but for which an
      electrode number is requested electrode positions for given electrode
      numbers
    * [done] need proper tests for all sorters
    * [done] how to deal with incorrect electrode positions, e.g. for the case
      where default Syscal settings are used?
    * [done] Merge two datasets by their electrode coordinates:


## User Stories

1) Import Syscal Data with unused electrodes but correct electrode positions

Syscal data contains only electrode positions (x,y,z for binary, but often only
x coordinates for ascii export). As such we need to assign electrode numbers to
those coordinates. This is relatively straightforward in case all electrodes
attached to the multichannel system were used and a regular spacing was set in
the system. However, if electrodes were skipped, things can get messy.

Two possible procedures can now be employed:

A)

    1) Add all encountered electrode positions (for a,b,m,n) to the electrode
       manager.
    2) Sort the positions in ascending order for z,y,x coordinates (in
       ascending priority)
    3) (optional) add missing electrode positions to conform the resulting
       assignments to a fixed grid of electrodes, e.g., 48, 72, or 96
       electrodes.
    4) Use the resulting electrode number assignment to generate logical
       electrode numbers for a,b,m,n

B)

    1) Provide a precomputed assignment table for electrode numbers and
       positions to the electrode manager
    2) Use the resulting electrode number assignment to generate logical
       electrode numbers for a,b,m,n


2) Import Syscal Data with unused electrodes but incorrect electrode positions

Same as in (1), but use a replacement function to replace the incorrect
electrode positions with the correct ones.

"""
import pandas as pd
import numpy as np


def decorator_name_index(func):
    def rename_index(ref, *args, **kwargs):
        result = func(ref)
        result.index.name = 'electrode_number'
        return result
    return rename_index


class electrode_manager(object):

    def __init__(self, electrode_positions=None):
        self.round_to_decimals = 6

        # overview of built-in ordering schemes
        self.ordering_schemes = {
            'as_is': self._order_as_is,
            'as_is_plus_one': self._order_as_is_plus_one,
            'sort_coordinates_zyx': self._order_coords_ascending_xyz
        }

        # if this is True, then we cannot change the electrode positions any
        # more. It used in case the user supplies her own assignment table
        # self.is_locked_down = False
        self.fixed_assigment_table = None

        if electrode_positions is None:
            self._electrode_positions = pd.DataFrame()
        else:
            self._electrode_positions = electrode_positions
        self.sorter = self.ordering_schemes['as_is_plus_one']

    def _order_coords_ascending_xyz(self, subdata):
        """Assign electrode coordinates by sorting using increasing
        coordinates, with decreasing priority of z, y, x axes.

        Adjust the index to start at 1.

        Parameters
        ----------
        subdata : pandas.DataFrame
            Data to order (columns x,y,z)

        Returns
        -------
        coordinates_sorted : pandas.DataFrame
            Ordered columns x, y, z
        """
        coordinates_sorted = subdata.sort_values(['z', 'y', 'x'])
        coordinates_sorted.index = range(1, coordinates_sorted.shape[0] + 1)
        return coordinates_sorted

    def _order_as_is(self, subdata):
        """Assign electrode numbers to coordinates in the order they were added

        Adjust the index to start at 1.

        Parameters
        ----------
        subdata : pandas.DataFrame
            Data to order (columns x,y,z)

        Returns
        -------
        coordinates_sorted : pandas.DataFrame
            Ordered columns x, y, z
        """
        coordinates_ordered = subdata.copy()

        return coordinates_ordered

    def _order_as_is_plus_one(self, subdata):
        """Assign electrode numbers to coordinates in the order they were added

        Adjust the index to start at 1.

        Parameters
        ----------
        subdata : pandas.DataFrame
            Data to order (columns x,y,z)

        Returns
        -------
        coordinates_sorted : pandas.DataFrame
            Ordered columns x, y, z
        """
        coordinates_ordered = subdata.copy()
        coordinates_ordered.index = range(1, coordinates_ordered.shape[0] + 1)

        return coordinates_ordered

    def register_custom_ordering_scheme(self, function):
        """Register a user-supplied function to assign electrode numbers to a
        given set of coordinates.

        The function must take one argument, a pandas.DataFrame with electrode
        coordinates in columns x,y,z. It then returns a ordered copy of that
        DataFrame, whose index values denote the electrode numbers. It is
        recommended to adjust the index to start with electrode number 1.
        """
        self.sorter = function

    def set_ordering_to_sort_zyx(self):
        """Helper function that activates the sorter that actually sorts
        electrode coordinates before assigning the electrode numbers.
        Useful to assign electrode numbers for automatically generated
        electrode numbers, such as encountered for some measurement systems.
        """
        self.sorter = self.ordering_schemes['sort_coordinates_zyx']

    def set_ordering_to_as_is(self):
        """Helper function that activates the sorter that assigns electrode
        numbers in the order the coordinates were registered. Useful if
        coordinates are input by hand.
        """
        self.sorter = self.ordering_schemes['as_is']

    def set_ordering_to_as_is_plus_one(self):
        """Helper function that activates the sorter that assigns electrode
        numbers in the order the coordinates were registered. Useful if
        coordinates are input by hand.

        This function will add 1 to the index values to ensure we do not zero
        index.
        """
        self.sorter = self.ordering_schemes['as_is_plus_one']

    def add_fixed_assignments(self, data_raw):
        """Use a pre-determined electrode numbering scheme.

        This is accomplished by the following steps:

        1) clean all existing electrode positions already registered
        2) set a lock that prevents the addition of additional
        """
        self.set_ordering_to_as_is()
        assert isinstance(data_raw, pd.DataFrame)
        assert 'electrode_number' in data_raw.columns or \
            data_raw.index.name == 'electrode_number', \
            'Column electrode_number must be present as column or named index'
        assert 'x' in data_raw.columns, 'Column x must be present'
        assert 'y' in data_raw.columns, 'Column y must be present'
        assert 'z' in data_raw.columns, 'Column z must be present'

        if data_raw.index.name == 'electrode_number':
            self.fixed_assigment_table = data_raw
        else:
            self.fixed_assigment_table = data_raw.set_index('electrode_number')

        def fixed_sorter(x):
            return self.fixed_assigment_table

        self.sorter = fixed_sorter

    def add_by_position(self, data_raw, remove_duplicates=True, **kwargs):
        """Add electrodes by using only their positions (1D/2D/3D).

        Electrode positions are rounded to the self.round_to_decimals decimal
        to ensure we can do proper comparison and identification with the
        floats.

        Multiple formats are possible:

        1) pandas.DataFrame with columns (x/z) or (x/y/z)
        2) list (one position) or nested list (multiple positions)
        3) numpy array

        WARNING: For lists/tuples/arrays the number of supplied dimensions is
        determined by the second dimension after applying::

            nr_dimensions = np.atleast_2d(input_data).shape[1]

        As such, make sure to properly format 1d arrays!

        Parameters
        ----------
        data_raw : numpy.ndarray|tuple|list
            Coordinate(s) for the electrodes to add to the pool
        remove_duplicates : bool, optional
            If True, only add coordinates not already registered

        """
        if isinstance(data_raw, pd.DataFrame):
            data = data_raw
        elif isinstance(data_raw, (np.ndarray, list, tuple)):
            data = np.atleast_1d(data_raw)
            if len(data.shape) == 1:
                data = data[:, np.newaxis]
            data = pd.DataFrame(data)
            if data.shape[1] == 1:
                # assume only x coordinate
                data.columns = ['x', ]
            if data.shape[1] == 2:
                # assume only x/z coordinates
                data.columns = ['x', 'z']
            elif data.shape[1] == 3:
                data.columns = ['x', 'y', 'z']

        assert 'x' in data, 'column x must be present'
        subdata = data.copy()
        if 'y' not in subdata.columns:
            subdata['y'] = 0
        if 'z' not in subdata.columns:
            subdata['z'] = 0

        # make sure we have the correct order of columns before working with
        # numpy
        subdata = subdata.reindex(
            columns=['x', 'y', 'z']
        ).astype(float).round(decimals=self.round_to_decimals)

        data_pre = None
        if self._electrode_positions.size > 0:
            # add existing data
            data_pre = np.atleast_2d(self._electrode_positions.values)

        # loop through new coordinates
        for _, row in subdata.iterrows():
            if data_pre is None:
                data_pre = np.atleast_2d(row.values)
            else:
                already_there = np.where(
                    np.all(
                        np.isclose(data_pre, row.values),
                        axis=1
                    )
                )[0].size
                if kwargs.get('debug', False):
                    print('----------------------')
                    print(data_pre)
                    print('row', row.values)
                    print(data_pre == row.values)
                    print(np.isclose(data_pre, row.values))
                    print(
                        np.where(
                            np.all(
                                data_pre == row.values,
                                axis=1
                            )
                        )
                    )
                    print('--------------------')

                if already_there == 0:
                    data_pre = np.vstack(
                        (
                            data_pre,
                            row.values,
                        )
                    )

        self._electrode_positions = pd.DataFrame(
            data_pre,
            columns=['x', 'y', 'z'],
        )

    def get_all_electrode_numbers(self):
        """Return the assignment between electrode numbers and coordinates
        """
        return self.electrode_positions

    @property
    @decorator_name_index
    def electrode_positions(self):
        return self.sorter(self._electrode_positions)

    @decorator_name_index
    def __call__(self):
        return self.sorter(self._electrode_positions)

    def get_electrode_numbers_for_positions(self, positions_raw):
        """For a given set of coordinates, return electrode coordinates

        In order to prevent problems with floating point representation and
        comparison, positions are rounded to the decimal given in
        self.round_to_decimals.

        Returns
        -------
        position : numpy.ndarray|None
        """

        assert isinstance(
            positions_raw, (tuple, list, np.ndarray, pd.DataFrame)
        ), 'Positions must be given as lists, tuples, numpy arrays, or' + \
            'a pandas.DataFrame'
        if isinstance(positions_raw, (list, tuple, np.ndarray)):
            positions_np = np.atleast_2d(positions_raw)
            if positions_np.shape[1] == 1:
                positions_np = np.vstack(
                    (
                        positions_np[:, 0],
                        np.zeros(positions_np.shape[0]),
                        np.zeros(positions_np.shape[0]),
                    )).T

            if positions_np.shape[1] == 2:
                positions_np = np.vstack(
                    (
                        positions_np[:, 0],
                        np.zeros(positions_np.shape[0]),
                        positions_np[:, 1]
                    )).T

            positions = pd.DataFrame(
                positions_np,
                columns=['x', 'y', 'z']
            )
        else:
            positions = positions_raw

        positions = positions.round(decimals=self.round_to_decimals)

        position = positions.merge(
            self.electrode_positions.reset_index(),
            on=['x', 'y', 'z'],
            how='left',
        )['electrode_number'].values
        # import IPython
        # IPython.embed()
        if position.size == 1 and np.isnan(position[0]):
            position = None
        return position

    def reflect_on_position_x(self, reflect_on_x):
        """Reflect all electrode positions on a given x position::

            x_new = reflect_on_x - x_i

        This procedure modifies the actual position data and cannot be
        reversed.

        It is useful to treat reciprocal data measured with devices that can be
        configured to measure reciprocal data easily, while retaining the
        electrode coordinates of the normal spread (e.g., possible with the
        Iris Instruments Syscal devices.
        """
        self._electrode_positions[
            'x'
        ] = reflect_on_x - self._electrode_positions['x']

        self._electrode_positions.round(decimals=self.round_to_decimals)

    def conform_to_regular_x_spacing(
            self, spacing_x, nr_electrodes=None, y_level=0, z_level=0):
        """Assume electrodes are located on a regular grid of electrodes with
        spacing spacing_x. Add missing electrode positions and thus ensure that
        the resulting electrode numbers conform to this full set of electrodes.

        This is useful when dealing with multiple measurements using
        multi-channel/multiplexer systems where not all electrodes are used in
        every measurement.

        This function works for simple setups - for specialized cases it is
        recommended to provide a complete electrode position list using the XXX
        method.

        Electrode positions are assumed to start at 0.

        Parameters
        ----------
        spacing_x : float
            Spacing of electrode meash
        nr_electrodes : int|None
            If set, only fill in electrode positions up to this electrode.
            Otherwise the maximal x position is used to determine the electrode
            number.
        y_level : float, optional
            y coordinate used for the additional points. Defaults to 0
        z_level : float, optional
            z coordinate used for the additional points. Defaults to 0

        Raises
        ------
        Exception in case where existing electrode positions are not multiples
        of spacing_x.
        """
        # check if all existing positions fit
        if np.any(self._electrode_positions['x'] % spacing_x > 0):
            raise Exception('spacing_x does not fit to all electrodes')

        if nr_electrodes is not None:
            max_x = (nr_electrodes - 1) * spacing_x + spacing_x
        else:
            max_x = self._electrode_positions['x'].max()

        new_positions = pd.DataFrame()
        new_positions['x'] = np.arange(0, max_x, step=spacing_x).round(
            decimals=self.round_to_decimals)
        new_positions['y'] = y_level
        new_positions['z'] = z_level

        self.add_by_position(new_positions)

    def get_position_of_number(self, number):
        """Return the position associated with a given electrode number.

        """
        assert isinstance(number, int), 'electrode numbers must be integers'
        electrode_positions = self.electrode_positions
        assert number in electrode_positions.index, \
            'electrode number not assigned yet'
        return self.electrode_positions.loc[number]

    def replace_coordinates_with_fixed_regular_grid_x(self, spacing_x):
        """Using the current ordering scheme, replace the x coordinates with
        multiples of spaxing_x, starting with 0.

        Parameters
        ----------
        spacing_x : float
            New spacing of x coordinates

        """
        old_coordinates = self.electrode_positions

        new_x_coordinates = np.arange(
            0, old_coordinates.shape[0] * spacing_x, spacing_x)

        old_coordinates['x'] = new_x_coordinates

        self._electrode_positions = old_coordinates

    def replace_coordinate_of_electrode_number(
            self, electrode_number, coordinates):
        """Replace the coordinates of the given electrode number (in the
        current ordering) with the new coordinates.

        Note that, depending on the chosen ordering scheme this can lead to the
        electrode assignments changing. This is the case for the
        'sort_coordinates_zyx' scheme. However, 'as_is' and 'as_is_plus_one'
        will retain the electrode order.

        Parameters
        ----------
        electrode_number : int
            The electrode number to replace
        coordinates : tuple|list|numpy.ndarray of size 2 or 3
            The new coordinates
        """
        assert isinstance(
            electrode_number, int), 'electrode number must be an integer'

        assert len(coordinates) in (2, 3), \
            'coordinates must be of length 2 or 3'

        if len(coordinates) == 2:
            coordinates_3d = [coordinates[0], 0, coordinates[1]]
        else:
            coordinates_3d = coordinates

        new_coords = np.array(coordinates_3d)

        current_positions = self.electrode_positions

        assert electrode_number in current_positions.index.values, \
            'Electrode number not registered yet'

        old_coordinates = self.electrode_positions.loc[electrode_number]

        index = np.where(
            np.all(
                old_coordinates == self._electrode_positions,
                axis=1
            )
        )
        self._electrode_positions.iloc[index] = new_coords.round(
            decimals=self.round_to_decimals)

    def plot_coordinates_x_z_to_ax(
            self, ax, plot_electrode_numbers=True, use_y_axis=False):
        coordinates = self.electrode_positions
        x = coordinates['x']
        if use_y_axis:
            y = coordinates['y']
        else:
            y = coordinates['z']

        ax.scatter(
            x,
            y,
            color='k',
        )
        if plot_electrode_numbers:
            for electrode_number, xyz in coordinates.iterrows():
                if use_y_axis:
                    y = xyz['y']
                else:
                    y = xyz['z']
                ax.text(
                    xyz['x'],
                    y,
                    '{}'.format(electrode_number),
                    fontsize=7,
                    bbox=dict(boxstyle='circle', facecolor='red', alpha=0.8)
                )

    def align_assignments(self, pos1, pos2, abmn1, abmn2):
        """Align the positions and logical electrode numbers of two datasets.

        Parameters
        ----------
        pos1 : pandas.DataFrame
            Dataframe containing the electrode positions of the first dataset.
            Required columns: x, y, z and either a column 'electrode_number' or
            an index named 'electrode_number'.
        pos2 : pandas.DataFrame
            Dataframe containing the electrode positions of the second dataset.
            Required columns: x, y, z and either a column 'electrode_number' or
            an index named 'electrode_number'.
        abmn1 : pandas.DataFrame
            Dataframe containing the logical electrode numbers a,b,m,n for the
            data set 1.
        abmn2 : pandas.DataFrame
            Dataframe containing the logical electrode numbers a,b,m,n for the
            data set 2.

        Returns
        -------
        electrode_positions_aligned : pandas.DataFrame
            Aligned electrode positions with new electrode numbers (index is
            named 'electrode_number').
        abmn1_aligned  : pandas.DataFrame
            Dataframe containing the logical electrode numbers a,b,m,n of the
            first data set after merging (aligning) the electrode positions of
            both datasets.
        abmn2_aligned  : pandas.DataFrame
            Dataframe containing the logical electrode numbers a,b,m,n of the
            second data set after merging (aligning) the electrode positions of
            both datasets.
        """
        assert isinstance(pos1, pd.DataFrame)
        assert isinstance(pos2, pd.DataFrame)
        assert isinstance(abmn1, pd.DataFrame)
        assert isinstance(abmn2, pd.DataFrame)
        assert 'x' in pos1.columns
        assert 'y' in pos1.columns
        assert 'z' in pos1.columns
        assert 'x' in pos2.columns
        assert 'y' in pos2.columns
        assert 'z' in pos2.columns

        for col in ['a', 'b', 'm', 'n']:
            for pos in (abmn1, abmn2):
                assert col in pos.columns

        pos1_rounded = pos1.round(decimals=self.round_to_decimals)
        pos2_rounded = pos2.round(decimals=self.round_to_decimals)

        elecs = electrode_manager()
        elecs.add_by_position(pos1_rounded)
        elecs.add_by_position(pos2_rounded)
        # print('merging here')
        # import IPython
        # IPython.embed()

        if pos1_rounded.index.name == 'electrode_number':
            pos1_rounded = pos1_rounded.reset_index()

        if pos2_rounded.index.name == 'electrode_number':
            pos2_rounded = pos2_rounded.reset_index()

        replacement_table1 = pd.merge(
            pos1_rounded,
            elecs().reset_index(), on=['x', 'y', 'z'], how='left')
        replacement_table2 = pd.merge(
            pos2_rounded,
            elecs().reset_index(), on=['x', 'y', 'z'], how='left')

        abmn1_aligned = abmn1.replace(
            replacement_table1['electrode_number_x'].values,
            replacement_table1['electrode_number_y'].values
        )

        abmn2_aligned = abmn2.replace(
            replacement_table2['electrode_number_x'].values,
            replacement_table2['electrode_number_y'].values
        )

        return elecs(), abmn1_aligned, abmn2_aligned

        # import IPython
        # IPython.embed()

    def shift_positions_xyz(self, shift_by_xyz):
        """Shift electrode positions by adding the vector 'shift_by_xyz' to
        each position.

        Parameters
        ----------
        shift_by_xyz : tuple|list|numpy.ndarray of size 1 or 2 or 3
            The vector to shift the data by. Length of 1 assumes that only the
            x coordinate of the vector differs from zero. Length of 2 assume a
            shift in (x,z) direction.
        """
        shift = np.atleast_1d(np.squeeze(np.array(shift_by_xyz)))
        assert len(shift.shape) == 1, 'only one 1D array is accepted'

        if shift.size == 1:
            shift = np.hstack((shift, 0, 0))
        elif shift.size == 2:
            shift = np.hstack((shift[0], 0, shift[1]))
        assert shift.size == 3, 'only arrays of length 1/2/3 are allowed'
        self._electrode_positions[['x', 'y', 'z']] += shift



"""Complex-resistivity container
"""
import os
import functools
import pandas as pd

import reda.utils.mpl

from reda.containers.BaseContainer import ImportersBase
from reda.containers.BaseContainer import BaseContainer

import reda.importers.bert as reda_bert_import
import reda.importers.mpt_das1 as reda_mpt
from reda.importers.crtomo import load_mod_file

from reda.utils.decorators_and_managers import append_doc_of
from reda.utils.decorators_and_managers import LogDataChanges

plt, mpl = reda.utils.mpl.setup()


class CRImporters(ImportersBase):
    def import_crtomo_data(self, filename):

        """
        Import a CRTomo-style measurement file (usually: volt.dat).

        Parameters
        ----------
        filename : str
            path to data file

        """

        data = load_mod_file(filename)
        self._add_to_container(data)

    @append_doc_of(reda_bert_import.import_ohm)
    def import_bert(self, filename, **kwargs):
        """BERT .ohm file import"""
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])

        self.logger.info('Unified data format (BERT/pyGIMLi) file import')
        with LogDataChanges(self, filter_action='import',
                            filter_query=os.path.basename(filename)):
            data, electrodes, topography = reda_bert_import.import_ohm(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)
            self.electrode_positions = electrodes  # See issue #22
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @append_doc_of(reda_mpt.import_das1_td)
    def import_mpt(self, filename, **kwargs):
        """
        MPT DAS 1 FD importer

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """

        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('MPT DAS-1 import')
        with LogDataChanges(self, filter_action='import'):
            data, electrodes, topography = reda_mpt.import_das1(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            if 'frequency' in data.columns:
                data = data.query(
                    'frequency == {}'.format(data.frequency.min())
                )
                data = data.drop('frequency')
            self._add_to_container(data)

        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @functools.wraps(import_bert)
    def import_pygimli(self, *args, **kargs):
        self.import_bert(*args, **kargs)


class CR(BaseContainer, CRImporters):
    """."""

    def __init__(self, data=None, electrode_positions=None, topography=None):
        """
        Parameters
        ----------
        data : :py:class:`pandas.DataFrame`
            If not None, then the provided DataFrame is assumed to contain
            valid data previously prepared elsewhere. Please refer to the
            documentation for required columns.
        electrode_positions : :py:class:`pandas.DataFrame`
            If set, this is expected to be a DataFrame which contains electrode
            positions with columns: "x", "y", "z".
        topography : :py:class:`pandas.DataFrame`
            If set, this is expected to a DataFrame which contains topography
            information with columns: "x", "y", "z".

        """
        self.setup_logger()
        self.required_columns = ['a',
                                 'b',
                                 'm',
                                 'n',
                                 'r',
                                 'rpha',
                                 'Zt']
        self.data = self.check_dataframe(data)
        self.electrode_positions = electrode_positions
        self.topography = topography

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required type and columns
        """
        if dataframe is None:
            return None

        # is this a DataFrame
        if not isinstance(dataframe, pd.DataFrame):
            raise Exception(
                'The provided dataframe object is not a pandas.DataFrame'
            )
        for column in self.required_columns:
            if column not in dataframe:
                raise Exception('Required column not in dataframe: {0}'.format(
                    column
                ))
        return dataframe

    def export_to_crtomo_td_manager(self, grid, norrec='norrec'):
        """Return a ready-initialized tdman object from the CRTomo tools.

        WARNING: Not timestep aware!

        Parameters
        ----------
        grid : crtomo.crt_grid
            A CRTomo grid instance
        norrec : str (nor|rec|norrec)
            Which data to export. Default: norrec (all)
        """
        if norrec in ('nor', 'rec'):
            subdata = self.data.query('norrec == "{}"'.format(norrec))
        else:
            subdata = self.data
        import crtomo
        data = subdata[['a', 'b', 'm', 'n', 'r', 'rpha']]
        tdman = crtomo.tdMan(grid=grid, volt_data=data)
        return tdman



"""
Set up some logging facilities for reda

We differentiate multiple logging types/targets

1) general logging (using the python logging module)
2) a data processing log (which also uses the logging module), available via
   the container-inherent .logger object
"""
import logging
import datetime

logging.basicConfig(
    # filename='logfile.log',
    level=logging.WARNING,
    format='{asctime} - {name:<30} - %{levelname:<10} - {message}',
    style='{',
)


class ListHandler(logging.Handler):  # Inherit from logging.Handler
    def __init__(self, log_list):
        # run the regular Handler __init__
        logging.Handler.__init__(self)
        # Our custom argument
        self.log_list = log_list

    def emit(self, record):
        # record.message is the log message
        self.log_list.append(record)


class LoggingClass(object):
    """Set up logging facilities for the containers
    """

    def setup_logger(self, name=None):
        """Setup a logger

        Parameters
        ----------
        name : str, optional
            If set, use this as the name of the logger
        """
        self.log_list = []
        handler = ListHandler(self.log_list)

        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        handler.setFormatter(formatter)

        if name is None:
            name = __name__
        logger = logging.getLogger(name)
        logger.addHandler(handler)

        logger.setLevel(logging.INFO)

        self.handler = handler
        self.logger = logger

    def print_log(self):
        for record in self.log_list:
            print(self.handler.format(record))

    def print_data_journal(self):
        print('')
        print('--- Data Journal Start ---')
        print('{0}'.format(datetime.datetime.now()))
        for record in self.log_list:
            if hasattr(record, 'filter_action'):
                # print(record)
                if record.filter_action == 'import':
                    print('Data was imported from file {0} '.format(
                        record.filter_query) + '({0} data points)'.format(
                            record.df_size_after - record.df_size_before))
                if record.filter_action == 'filter':
                    print(
                        'A filter was applied with query "{0}".'.format(
                            record.filter_query
                        ) +
                        ' In total {0} records were removed'.format(
                            -record.df_size_after + record.df_size_before
                        )
                    )
        print('--- Data Journal End ---')
        print('')



# -*- coding: utf-8 -*-
""" Work with result files from the EIT-40/160 tomograph (also called medusa).

Data structure of .mat files:

    EMD(n).fm frequency
    EMD(n).Time point of time of this measurement
    EMD(n).ni number of the two excitation electrodes (not the channel number)
    EMD(n).nu number of the two potential electrodes (not the channel number)
    EMD(n).Zt3 array with the transfer impedances (repetition measurement)
    EMD(n).nni number of injection
    EMD(n).cni number of channels used to inject current
    EMD(n).cnu number of channels used to measure voltage
    EMD(n).Is3 injected current (A) (calibrated)
    EMD(n).II3 leakage current (A)
    EMD(n).Yg1 Admitance of first injection path
    EMD(n).Yg2 Admitance of second injection path
    EMD(n).As3 Voltages at shunt resistors (defined in .mcf files: NA1 - NA2)
    EMD(n).Zg3 Impedance between injection electrodes


Import pipeline:

* read single-potentials from .mat file
* read quadrupoles from separate file or provide numpy array
* compute mean of three impedance measurement repetitions (Z1-Z3) for each ABM
* compute quadrupole impedance via superposition using

    * a) averaged Z-values
    * b) the single repetitions Z1-Z3

* (I think we don't need the next step because of np.arctan2)
  check for correct quadrant in phase values, correct if necessary (is this
  required if we use the arctan2 function?)
* compute variance/standard deviation from the repetition values

Open questions:

* should we provide a time delta between the two measurements?

"""
print('WARNING: DONt use this module anymore!. Use the eit_fzj module!')
import numpy as np
import scipy.io as sio
import pandas as pd
import datetime

import reda.utils.geometric_factors as redaK


def _add_rhoa(df, spacing):
    """a simple wrapper to compute K factors and add rhoa
    """
    df['k'] = redaK.compute_K_analytical(df, spacing=spacing)
    df['rho_a'] = df['r'] * df['k']
    if 'Zt' in df.columns:
        df['rho_a_complex'] = df['Zt'] * df['k']
    return df


def import_medusa_data(mat_filename, config_file):
    """Import measurement data (a .mat file) of the FZJ EIT160 system. This
    data format is identified as 'FZJ-EZ-2017'.

    Parameters
    ----------
    mat_filename: string
        filename to the .mat data file. Note that only MNU0 single-potentials
        are supported!
    config_file: string
        filename for configuration file. The configuration file contains N rows
        with 4 columns each (a, b, m, n)

    Returns
    -------

    """
    df_emd, df_md = _read_mat_mnu0(mat_filename)

    # 'configs' can be a numpy array or a filename
    if not isinstance(config_file, np.ndarray):
        configs = np.loadtxt(config_file).astype(int)
    else:
        configs = config_file

    # construct four-point measurements via superposition
    print('constructing four-point measurements')
    quadpole_list = []

    if df_emd is not None:
        index = 0
        for Ar, Br, M, N in configs:
            # print('constructing', Ar, Br, M, N)
            # the order of A and B doesn't concern us
            A = np.min((Ar, Br))
            B = np.max((Ar, Br))

            # first choice: correct ordering
            query_M = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
                A, B, M
            ))
            query_N = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
                A, B, N
            ))

            if query_M.size == 0 or query_N.size == 0:
                continue

            index += 1

            # keep these columns as they are (no subtracting)
            keep_cols = [
                'datetime',
                'frequency',
                'a', 'b',
                'Zg1', 'Zg2', 'Zg3',
                'Is',
                'Il',
                'Zg',
                'Iab',
            ]

            df4 = pd.DataFrame()
            diff_cols = ['Zt', ]
            df4[keep_cols] = query_M[keep_cols]
            for col in diff_cols:
                df4[col] = query_M[col].values - query_N[col].values
            df4['m'] = query_M['p'].values
            df4['n'] = query_N['p'].values

            quadpole_list.append(df4)

    if quadpole_list:
        dfn = pd.concat(quadpole_list)
        Rsign = np.sign(dfn['Zt'].real)
        dfn['r'] = Rsign * np.abs(dfn['Zt'])
        dfn['Vmn'] = dfn['r'] * dfn['Iab']
        dfn['rpha'] = np.arctan2(
            np.imag(dfn['Zt'].values),
            np.real(dfn['Zt'].values)
        ) * 1e3
    else:
        dfn = pd.DataFrame()

    return dfn, df_md


def _read_mat_mnu0(filename):
    """Import a .mat file with single potentials (a b m) into a pandas
    DataFrame

    Also export some variables of the MD struct into a separate structure
    """
    print('read_mag_single_file: {0}'.format(filename))

    mat = sio.loadmat(filename, squeeze_me=True)
    # check the version
    version = mat['MP']['Version'].item()
    if version != 'FZJ-EZ-2017':
        raise Exception(
            'This data format is not supported (expected: FZJ-EZ-2017)' +
            ' got: {}'.format(version)
        )

    df_emd = _extract_emd(mat, filename=filename)
    df_md = _extract_md(mat)

    return df_emd, df_md


def _extract_md(mat):
    md = mat['MD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, md.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = md[f_id]
        # for name in fdata.dtype.names:
        #     print(name, fdata[name].shape)

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['cni'],
                fdata['Cl3'],
                fdata['Zg3'],
                fdata['As3'][:, 0, :].squeeze(),
                fdata['As3'][:, 1, :].squeeze(),
                fdata['As3'][:, 2, :].squeeze(),
                fdata['As3'][:, 3, :].squeeze(),
                fdata['Is3'],
                fdata['Yl3'],
            ))
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'Cl1',
            'Cl2',
            'Cl3',
            'Zg1',
            'Zg2',
            'Zg3',
            'ShuntVoltage1_1',
            'ShuntVoltage1_2',
            'ShuntVoltage1_3',
            'ShuntVoltage2_1',
            'ShuntVoltage2_2',
            'ShuntVoltage2_3',
            'ShuntVoltage3_1',
            'ShuntVoltage3_2',
            'ShuntVoltage3_3',
            'ShuntVoltage4_1',
            'ShuntVoltage4_2',
            'ShuntVoltage4_3',
            'Is1',
            'Is2',
            'Is3',
            'Yl1',
            'Yl2',
            'Yl3',
        )

        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['Cl1'] = df['Cl1'].astype(complex)
        df['Cl2'] = df['Cl2'].astype(complex)
        df['Cl3'] = df['Cl3'].astype(complex)
        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Yl1'] = df['Yl1'].astype(complex)
        df['Yl2'] = df['Yl2'].astype(complex)
        df['Yl3'] = df['Yl3'].astype(complex)

        df['ShuntVoltage1_1'] = df['ShuntVoltage1_1'].astype(complex)
        df['ShuntVoltage1_2'] = df['ShuntVoltage1_2'].astype(complex)
        df['ShuntVoltage1_3'] = df['ShuntVoltage1_3'].astype(complex)

        df['ShuntVoltage2_1'] = df['ShuntVoltage2_1'].astype(complex)
        df['ShuntVoltage2_2'] = df['ShuntVoltage2_2'].astype(complex)
        df['ShuntVoltage2_3'] = df['ShuntVoltage2_3'].astype(complex)

        df['ShuntVoltage3_1'] = df['ShuntVoltage3_1'].astype(complex)
        df['ShuntVoltage3_2'] = df['ShuntVoltage3_2'].astype(complex)
        df['ShuntVoltage3_3'] = df['ShuntVoltage3_3'].astype(complex)

        df['ShuntVoltage4_1'] = df['ShuntVoltage4_1'].astype(complex)
        df['ShuntVoltage4_2'] = df['ShuntVoltage4_2'].astype(complex)
        df['ShuntVoltage4_3'] = df['ShuntVoltage4_3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
        # "standard" injected current, in [mA]
        df['Iab'] = np.abs(df['Is']) * 1e3
        df['Iab'] = df['Iab'].astype(float)

        df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']], axis=1)

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']
        dfl.append(df)

    df = pd.concat(dfl)

    return df


def _extract_emd(mat, filename):
    """Extract the data from the EMD substruct, given a medusa-created MNU0-mat
    file

    Parameters
    ----------

    mat: matlab-imported struct

    """
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # some consistency checks
        if len(fdata['nu']) == 2 and fdata['nu'].shape[1] == 2:
            raise Exception(
                'Need MNU0 file, not a quadpole .mat file: {0}'.format(
                    filename
                )
            )

        # fdata_md = md[f_id]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['ni'],
                fdata['nu'][:, np.newaxis],
                fdata['Zt3'],
                fdata['Is3'],
                fdata['Il3'],
                fdata['Zg3'],
                fdata['As3'][:, 0, :].squeeze(),
                fdata['As3'][:, 1, :].squeeze(),
                fdata['As3'][:, 2, :].squeeze(),
                fdata['As3'][:, 3, :].squeeze(),
                fdata['Yg13'],
                fdata['Yg23'],
            )),
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'p',
            'Z1',
            'Z2',
            'Z3',
            'Is1',
            'Is2',
            'Is3',
            'Il1',
            'Il2',
            'Il3',
            'Zg1',
            'Zg2',
            'Zg3',
            'ShuntVoltage1_1',
            'ShuntVoltage1_2',
            'ShuntVoltage1_3',
            'ShuntVoltage2_1',
            'ShuntVoltage2_2',
            'ShuntVoltage2_3',
            'ShuntVoltage3_1',
            'ShuntVoltage3_2',
            'ShuntVoltage3_3',
            'ShuntVoltage4_1',
            'ShuntVoltage4_2',
            'ShuntVoltage4_3',
            'Yg13_1',
            'Yg13_2',
            'Yg13_3',
            'Yg23_1',
            'Yg23_2',
            'Yg23_3',
        )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']

        # cast to correct type
        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['p'] = df['p'].astype(int)

        df['Z1'] = df['Z1'].astype(complex)
        df['Z2'] = df['Z2'].astype(complex)
        df['Z3'] = df['Z3'].astype(complex)

        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Il1'] = df['Il1'].astype(complex)
        df['Il2'] = df['Il2'].astype(complex)
        df['Il3'] = df['Il3'].astype(complex)

        df['ShuntVoltage1_1'] = df['ShuntVoltage1_1'].astype(complex)
        df['ShuntVoltage1_2'] = df['ShuntVoltage1_2'].astype(complex)
        df['ShuntVoltage1_3'] = df['ShuntVoltage1_3'].astype(complex)

        df['ShuntVoltage2_1'] = df['ShuntVoltage2_1'].astype(complex)
        df['ShuntVoltage2_2'] = df['ShuntVoltage2_2'].astype(complex)
        df['ShuntVoltage2_3'] = df['ShuntVoltage2_3'].astype(complex)

        df['ShuntVoltage3_1'] = df['ShuntVoltage3_1'].astype(complex)
        df['ShuntVoltage3_2'] = df['ShuntVoltage3_2'].astype(complex)
        df['ShuntVoltage3_3'] = df['ShuntVoltage3_3'].astype(complex)

        df['ShuntVoltage4_1'] = df['ShuntVoltage4_1'].astype(complex)
        df['ShuntVoltage4_2'] = df['ShuntVoltage4_2'].astype(complex)
        df['ShuntVoltage4_3'] = df['ShuntVoltage4_3'].astype(complex)

        dfl.append(df)

    if len(dfl) == 0:
        return None
    df = pd.concat(dfl)

    # average swapped current injections here!
    # TODO

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[condition, ['b', 'a']].values
    # change sign because we changed A and B
    df.loc[condition, ['Z1', 'Z2', 'Z3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Z1', 'Z2', 'Z3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = df['Zt'].real / np.abs(df['Zt'].real)
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    df['Iab'] = np.abs(df['Is']) * 1e3
    df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)

    return df



"""Sort configurations into four-point spread types.

Sorting is done by subsequently applying filters to the configurations, with
removal of selected configurations. Thus, each filter sees only configurations
not 'chosen' by any previously applied filters.

    * dipole-dipole
    * Schlumberger
    * Wenner
    * misc

"""
import numpy as np
import pandas as pd


def _filter_schlumberger(configs):
    """Filter Schlumberger configurations

    Schlumberger configurations are selected using the following criteria:

    * For a given voltage dipole, there need to be at least two current
      injections with electrodes located on the left and the right of the
      voltage dipole.
    * The distance between the current electrodes and the next voltage
      electrode is the same on both sides.

    Parameters
    ----------
    configs: numpy.ndarray
        Nx4 array with N measurement configurations

    Returns
    -------
    configs: numpy.ndarray
        Remaining configurations, all Schlumberger configurations are set to
        numpy.nan
    schl_indices: dict with one entry: numpy.ndarray
        indices of Schlumberger configurations
    """
    # sort configs
    configs_sorted = np.hstack((
        np.sort(configs[:, 0:2], axis=1),
        np.sort(configs[:, 2:4], axis=1),
    )).astype(int)

    # determine unique voltage dipoles
    MN = configs_sorted[:, 2:4].copy()
    MN_unique = np.unique(
        MN.view(
            MN.dtype.descr * 2
        )
    )
    MN_unique_reshape = MN_unique.view(
        MN.dtype
    ).reshape(-1, 2)

    schl_indices_list = []
    for mn in MN_unique_reshape:
        # check if there are more than one associated current injections
        nr_current_binary = (
            (configs_sorted[:, 2] == mn[0]) &
            (configs_sorted[:, 3] == mn[1])
        )
        if len(np.where(nr_current_binary)[0]) < 2:
            continue

        # now which of these configurations have current electrodes on both
        # sides of the voltage dipole
        nr_left_right = (
            (configs_sorted[:, 0] < mn[0]) &
            (configs_sorted[:, 1] > mn[0]) &
            nr_current_binary
        )

        # now check that the left/right distances are equal
        distance_left = np.abs(
            configs_sorted[nr_left_right, 0] - mn[0]
        ).squeeze()
        distance_right = np.abs(
            configs_sorted[nr_left_right, 1] - mn[1]
        ).squeeze()

        nr_equal_distances = np.where(distance_left == distance_right)[0]

        indices = np.where(nr_left_right)[0][nr_equal_distances]

        if indices.size > 2:
            schl_indices_list.append(indices)

    # set Schlumberger configs to nan
    if len(schl_indices_list) == 0:
        return configs, {0: np.array([])}
    else:
        schl_indices = np.hstack(schl_indices_list).squeeze()
        configs[schl_indices, :] = np.nan
        return configs, {0: schl_indices}


def _filter_dipole_dipole(configs):
    """Filter dipole-dipole configurations

    A dipole-dipole configuration is defined using the following criteria:

        * equal distance between the two current electrodes and between the two
          voltage electrodes
        * no overlap of dipoles

    Parameters
    ----------
    configs: numpy.ndarray
        Nx4 array with N measurement configurations

    Returns
    -------
    configs: numpy.ndarray
        Remaining configurations, all dipole-dipole configurations are set to
        numpy.nan
    dd_indices: numpy.ndarray
        indices of dipole-dipole configurations

    """
    # check that dipoles have equal size
    dist_ab = np.abs(configs[:, 0] - configs[:, 1])
    dist_mn = np.abs(configs[:, 2] - configs[:, 3])

    distances_equal = (dist_ab == dist_mn)

    # check that they are not overlapping
    not_overlapping = (
        # either a,b < m,n
        (
            (configs[:, 0] < configs[:, 2]) &
            (configs[:, 1] < configs[:, 2]) &
            (configs[:, 0] < configs[:, 3]) &
            (configs[:, 1] < configs[:, 3])
        ) |
        # or m,n < a,b
        (
            (configs[:, 2] < configs[:, 0]) &
            (configs[:, 3] < configs[:, 0]) &
            (configs[:, 2] < configs[:, 1]) &
            (configs[:, 3] < configs[:, 1])
        )
    )

    is_dipole_dipole = (distances_equal & not_overlapping)

    dd_indices = np.where(is_dipole_dipole)[0]
    dd_indices_sorted = _sort_dd_skips(configs[dd_indices, :], dd_indices)

    # set all dd configs to nan
    configs[dd_indices, :] = np.nan

    return configs, dd_indices_sorted


def _sort_dd_skips(configs, dd_indices_all):
    """Given a set of dipole-dipole configurations, sort them according to
    their current skip.

    Parameters
    ----------
    configs: Nx4 numpy.ndarray
        Dipole-Dipole configurations

    Returns
    -------
    dd_configs_sorted: dict
        dictionary with the skip as keys, and arrays/lists with indices to
        these skips.
    """
    config_current_skips = np.abs(configs[:, 1] - configs[:, 0])
    if np.all(np.isnan(config_current_skips)):
        return {0: []}

    # determine skips
    available_skips_raw = np.unique(config_current_skips)
    available_skips = available_skips_raw[
        ~np.isnan(available_skips_raw)
    ].astype(int)

    # now determine the configurations
    dd_configs_sorted = {}
    for skip in available_skips:
        indices = np.where(config_current_skips == skip)[0]
        dd_configs_sorted[skip - 1] = dd_indices_all[indices]

    return dd_configs_sorted


def filter(configs, settings):
    """Main entry function to filtering configuration types

    Parameters
    ----------
    configs: Nx4 array
        array containing A-B-M-N configurations
    settings: dict
        'only_types': ['dd', 'other'],  # filter only for those types

    Returns
    -------
    dict
        results dict containing filter results (indices) for all registered
        filter functions.  All remaining configs are stored under the keywords
        'remaining'

    """
    if isinstance(configs, pd.DataFrame):
        configs = configs[['a', 'b', 'm', 'n']].values

    # assign short labels to Python functions
    filter_funcs = {
        'dd': _filter_dipole_dipole,
        'schlumberger': _filter_schlumberger,
    }

    # we need a list to fix the call order of filter functions
    keys = ['dd', 'schlumberger', ]

    allowed_keys = settings.get('only_types', filter_funcs.keys())

    results = {}
    # we operate iteratively on the configs, set the first round here
    # rows are iteratively set to nan when filters remove them!
    configs_filtered = configs.copy().astype(float)

    for key in keys:
        if key in allowed_keys:
            configs_filtered, indices_filtered = filter_funcs[key](
                configs_filtered,
            )
            if len(indices_filtered) > 0:
                results[key] = indices_filtered

    # add all remaining indices to the results dict
    results['not_sorted'] = np.where(
        ~np.all(np.isnan(configs_filtered), axis=1)
    )[0]
    return results



"""2D plots of raw data. This includes pseudosections."""

import numpy as np
import scipy.interpolate as si
import scipy.spatial.qhull as siq

import reda.utils.filter_config_types as fT
import reda.utils.mpl

plt, mpl = reda.utils.mpl.setup()


def _pseudodepths_wenner(configs, spacing=1, grid=None):
    """Given distances between electrodes, compute Wenner pseudo
    depths for the provided configuration

    The pseudodepth is computed after Roy & Apparao, 1971, as 0.11 times
    the distance between the two outermost electrodes. It's not really
    clear why the Wenner depths are different from the Dipole-Dipole
    depths, given the fact that Wenner configurations are a complete subset
    of the Dipole-Dipole configurations.

    """
    if grid is None:
        xpositions = (configs - 1) * spacing
    else:
        xpositions = grid.get_electrode_positions()[configs - 1, 0]

    z = np.abs(np.max(xpositions, axis=1) - np.min(xpositions, axis=1)) * -0.11
    x = np.mean(xpositions, axis=1)
    return x, z


def _pseudodepths_schlumberger(configs, spacing=1, grid=None):
    """Given distances between electrodes, compute Schlumberger pseudo
    depths for the provided configuration

    The pseudodepth is computed after Roy & Apparao, 1971, as 0.125 times
    the distance between the two outermost electrodes.

    """
    if grid is None:
        xpositions = (configs - 1) * spacing
    else:
        xpositions = grid.get_electrode_positions()[configs - 1, 0]

    x = np.mean(xpositions, axis=1)
    z = np.abs(np.max(xpositions, axis=1) - np.min(xpositions,
                                                   axis=1)) * -0.125
    return x, z


def _pseudodepths_dd_simple(configs, spacing=1, grid=None):
    """Given distances between electrodes, compute dipole-dipole pseudo
    depths for the provided configuration

    The pseudodepth is computed after Roy & Apparao, 1971, as 0.195 times
    the distance between the two outermost electrodes.

    """
    if grid is None:
        xpositions = (configs - 1) * spacing
    else:
        xpositions = grid.get_electrode_positions()[configs - 1, 0]

    z = np.abs(np.max(xpositions, axis=1) - np.min(xpositions,
                                                   axis=1)) * -0.195
    x = np.mean(xpositions, axis=1)
    return x, z


def plot_pseudodepths(configs, nr_electrodes, spacing=1, grid=None,
                      ctypes=None, dd_merge=False, **kwargs):
    """Plot pseudodepths for the measurements. If grid is given, then the
    actual electrode positions are used, and the parameter 'spacing' is
    ignored'

    Parameters
    ----------
    configs: :class:`numpy.ndarray`
        Nx4 array containing the quadrupoles for different measurements
    nr_electrodes: int
        The overall number of electrodes of the dataset. This is used to plot
        the surface electrodes
    spacing: float, optional
        assumed distance between electrodes. Default=1
    grid: crtomo.grid.crt_grid instance, optional
        grid instance. Used to infer real electrode positions
    ctypes: list of strings, optional
        a list of configuration types that will be plotted. All
        configurations that can not be sorted into these types will not be
        plotted! Possible types:

        * dd
        * schlumberger

    dd_merge: bool, optional
        if True, merge all skips. Otherwise, generate individual plots for
        each skip

    Returns
    -------
    figs: matplotlib.figure.Figure instance or list of Figure instances
        if only one type was plotted, then the figure instance is returned.
        Otherwise, return a list of figure instances.
    axes: axes object or list of axes ojects
        plot axes

    Examples
    --------

    .. plot::
        :include-source:

        from reda.plotters.plots2d import plot_pseudodepths
        # define a few measurements
        import numpy as np
        configs = np.array((
            (1, 2, 4, 3),
            (1, 2, 5, 4),
            (1, 2, 6, 5),
            (2, 3, 5, 4),
            (2, 3, 6, 5),
            (3, 4, 6, 5),
        ))
        # plot
        fig, axes = plot_pseudodepths(configs, nr_electrodes=6, spacing=1,
                                      ctypes=['dd', ])

    .. plot::
        :include-source:

        from reda.plotters.plots2d import plot_pseudodepths
        # define a few measurements
        import numpy as np
        configs = np.array((
            (4, 7, 5, 6),
            (3, 8, 5, 6),
            (2, 9, 5, 6),
            (1, 10, 5, 6),
        ))
        # plot
        fig, axes = plot_pseudodepths(configs, nr_electrodes=10, spacing=1,
                                      ctypes=['schlumberger', ])

    """
    # for each configuration type we have different ways of computing
    # pseudodepths
    pseudo_d_functions = {
        'dd': _pseudodepths_dd_simple,
        'schlumberger': _pseudodepths_schlumberger,
        'wenner': _pseudodepths_wenner,
    }

    titles = {
        'dd': 'dipole-dipole configurations',
        'schlumberger': 'Schlumberger configurations',
        'wenner': 'Wenner configurations',
    }

    # sort the configurations into the various types of configurations
    only_types = ctypes or ['dd', ]
    results = fT.filter(configs, settings={'only_types': only_types, })

    # loop through all measurement types
    figs = []
    axes = []
    for key in sorted(results.keys()):
        print('plotting: ', key)
        if key == 'not_sorted':
            continue
        index_dict = results[key]
        # it is possible that we want to generate multiple plots for one
        # type of measurement, i.e., to separate skips of dipole-dipole
        # measurements. Therefore we generate two lists:
        # 1) list of list of indices to plot
        # 2) corresponding labels
        if key == 'dd' and not dd_merge:
            plot_list = []
            labels_add = []
            for skip in sorted(index_dict.keys()):
                plot_list.append(index_dict[skip])
                labels_add.append(' - skip {0}'.format(skip))
        else:
            # merge all indices
            plot_list = [np.hstack(index_dict.values()), ]
            print('schlumberger', plot_list)
            labels_add = ['', ]

        grid = None
        # generate plots
        for indices, label_add in zip(plot_list, labels_add):
            if len(indices) == 0:
                continue
            ddc = configs[indices]
            px, pz = pseudo_d_functions[key](ddc, spacing, grid)

            fig, ax = plt.subplots(figsize=(15 / 2.54, 5 / 2.54))
            ax.scatter(px, pz, color='k', alpha=0.5)

            # plot electrodes
            if grid is not None:
                electrodes = grid.get_electrode_positions()
                ax.scatter(
                    electrodes[:, 0],
                    electrodes[:, 1],
                    color='b',
                    label='electrodes', )
            else:
                ax.scatter(
                    np.arange(0, nr_electrodes) * spacing,
                    np.zeros(nr_electrodes),
                    color='b',
                    label='electrodes', )
            ax.set_title(titles[key] + label_add)
            ax.set_aspect('equal')
            ax.set_xlabel('x [m]')
            ax.set_ylabel('x [z]')

            fig.tight_layout()
            figs.append(fig)
            axes.append(ax)

    if len(figs) == 1:
        return figs[0], axes[0]
    else:
        return figs, axes


def plot_pseudosection(df, plot_key, spacing=1, ctypes=None, dd_merge=False,
                       cb=False, **kwargs):
    """Create a pseudosection plot for a given measurement

    Parameters
    ----------
    df: dataframe
        measurement dataframe, one measurement frame (i.e., only one frequency
        etc)
    key:
        which key to colorcode
    spacing: float, optional
        assumed electrode spacing
    ctypes: list of strings
        which configurations to plot, default: dd
    dd_merge: bool, optional
        ?
    cb: bool, optional
        ?

    """
    grid = None

    pseudo_d_functions = {
        'dd': _pseudodepths_dd_simple,
        'schlumberger': _pseudodepths_schlumberger,
        'wenner': _pseudodepths_wenner,
    }

    titles = {
        'dd': 'dipole-dipole configurations',
        'schlumberger': 'Schlumberger configurations',
        'wenner': 'Wenner configurations',
    }

    # for now sort data and only plot dipole-dipole
    only_types = ctypes or ['dd', ]
    if 'schlumberger' in only_types:
        raise Exception('plotting of pseudosections not implemented for ' +
                        'Schlumberger configurations!')

    configs = df[['a', 'b', 'm', 'n']].values
    results = fT.filter(
        configs,
        settings={'only_types': only_types, }, )
    values = df[plot_key].values

    plot_objects = []
    for key in sorted(results.keys()):
        print('plotting: ', key)
        if key == 'not_sorted':
            continue
        index_dict = results[key]
        # it is possible that we want to generate multiple plots for one
        # type of measurement, i.e., to separate skips of dipole-dipole
        # measurements. Therefore we generate two lists:
        # 1) list of list of indices to plot
        # 2) corresponding labels
        if key == 'dd' and not dd_merge:
            plot_list = []
            labels_add = []
            for skip in sorted(index_dict.keys()):
                plot_list.append(index_dict[skip])
                labels_add.append(' - skip {0}'.format(skip))
        else:
            # merge all indices
            plot_list = [np.hstack(index_dict.values()), ]
            print('schlumberger', plot_list)
            labels_add = ['', ]

        # generate plots
        for indices, label_add in zip(plot_list, labels_add):
            if len(indices) == 0:
                continue

            ddc = configs[indices]
            plot_data = values[indices]
            px, pz = pseudo_d_functions[key](ddc, spacing, grid)
            # we need at least four points for a spatial interpolation, I
            # think...
            if px.size <= 4:
                continue

            # take 200 points for the new grid in every direction. Could be
            # adapted to the actual ratio
            xg = np.linspace(px.min(), px.max(), 200)
            zg = np.linspace(pz.min(), pz.max(), 200)

            x, z = np.meshgrid(xg, zg)

            cmap_name = kwargs.get('cmap_name', 'jet')
            cmap = mpl.cm.get_cmap(cmap_name)

            # normalize data
            data_min = kwargs.get('cbmin', plot_data.min())
            data_max = kwargs.get('cbmax', plot_data.max())
            cnorm = mpl.colors.Normalize(vmin=data_min, vmax=data_max)
            scalarMap = mpl.cm.ScalarMappable(norm=cnorm, cmap=cmap)
            fcolors = scalarMap.to_rgba(plot_data)

            try:
                image = si.griddata(
                    (px, pz),
                    fcolors,
                    (x, z),
                    method='linear', )
            except siq.QhullError as e:
                print('Ex', e)
                continue

            cmap = mpl.cm.get_cmap('jet_r')

            data_ratio = np.abs(px.max() - px.min()) / np.abs(pz.min())

            fig_size_y = 15 / data_ratio + 6 / 2.54
            fig = plt.figure(figsize=(15, fig_size_y))

            fig_top = 1 / 2.54 / fig_size_y
            fig_left = 2 / 2.54 / 15
            fig_right = 1 / 2.54 / 15
            if cb:
                fig_bottom = 3 / 2.54 / fig_size_y
            else:
                fig_bottom = 0.05

            ax = fig.add_axes([
                fig_left, fig_bottom + fig_top * 2, 1 - fig_left - fig_right,
                1 - fig_top - fig_bottom - fig_top * 2
            ])

            im = ax.imshow(
                image[::-1],
                extent=(xg.min(), xg.max(), zg.min(), zg.max()),
                interpolation='none',
                aspect='auto',
                # vmin=10,
                # vmax=300,
                cmap=cmap, )
            ax.set_ylim(pz.min(), 0)

            # colorbar
            if cb:
                print('plotting colorbar')
                # the colorbar has 3 cm on the bottom
                ax_cb = fig.add_axes([
                    fig_left * 4, fig_top * 2,
                    1 - fig_left * 4 - fig_right * 4, fig_bottom - fig_top * 2
                ])
                # from mpl_toolkits.axes_grid1 import make_axes_locatable
                # divider = make_axes_locatable(ax)
                # ax_cb = divider.append_axes("bottom", "5%", pad="3%")
                # (ax_cb, kw) = mpl.colorbar.make_axes_gridspec(
                #     ax,
                #     orientation='horizontal',
                #     fraction=fig_bottom,
                #     pad=0.3,
                #     shrink=0.9,
                #     # location='bottom',
                # )
                cb = mpl.colorbar.ColorbarBase(
                    ax=ax_cb,
                    cmap=cmap,
                    norm=cnorm,
                    orientation='horizontal',
                    # **kw
                )
                cb.set_label('cb label')
            else:
                fig_bottom = 0.05

            # 1cm on top

            # # 3 cm on bottom for colorbar
            # fig.subplots_adjust(
            #     top=1 - fig_top,
            #     bottom=fig_bottom,
            # )

            ax.set_title(titles[key] + label_add)
            ax.set_aspect('equal')
            ax.set_xlabel('x [m]')
            ax.set_ylabel('x [z]')
            plot_objects.append((fig, ax, im))

    return plot_objects


def matplot(x, y, z, ax=None, colorbar=True, **kwargs):
    """ Plot x, y, z as expected with correct axis labels.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from reda.plotters import matplot
    >>> a = np.arange(4)
    >>> b = np.arange(3) + 3
    >>> def sum(a, b):
    ...    return a + b
    >>> x, y = np.meshgrid(a, b)
    >>> c = sum(x, y)
    >>> fig, (ax1, ax2) = plt.subplots(1, 2)
    >>> im = ax1.pcolormesh(x, y, c, shading='auto')
    >>> _ = plt.colorbar(im, ax=ax1)
    >>> _ = ax1.set_title("plt.pcolormesh")
    >>> _, _ = matplot(x, y, c, ax=ax2)
    >>> _ = ax2.set_title("reda.plotters.matplot")
    >>> fig.show()

    Note
    ----
    Only works for equidistant data at the moment.
    """
    xmin = x.min()
    xmax = x.max()
    dx = np.abs(x[0, 1] - x[0, 0])

    ymin = y.min()
    ymax = y.max()
    dy = np.abs(y[1, 0] - y[0, 0])

    x2, y2 = np.meshgrid(
        np.arange(xmin, xmax + 2 * dx, dx) - dx / 2.,
        np.arange(ymin, ymax + 2 * dy, dy) - dy / 2.)

    if not ax:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    im = ax.pcolormesh(
        x2, y2, z,
        shading='auto',
        **kwargs
    )
    ax.axis([x2.min(), x2.max(), y2.min(), y2.max()])
    ax.set_xticks(np.arange(xmin, xmax + dx, dx))
    ax.set_yticks(np.arange(ymin, ymax + dx, dy))

    if colorbar:
        cbar = fig.colorbar(im, ax=ax)
    else:
        cbar = None

    return ax, cbar


def plot_rawdataplot():
    pass



import reda.main.logger
from reda.utils.enter_directory import CreateEnterDirectory
from reda.utils.enter_directory import EnterDirectory
import reda.importers.utils.transforms as transforms
from .configs.configManager import ConfigManager
import reda.utils.data as data
from .utils.helper_functions import search
from .containers.BaseContainer import BaseContainer
from .containers.SIP import SIP
from .containers.sEIT import sEIT
from .containers.CR import CR
from .containers.TDIP import TDIP
from .containers.ERT import ERT
from .testing import test
import reda.importers.eit_fzj as eit_fzj
from reda.importers.fzj_readbin import fzj_readbin
from reda.utils.electrode_manager import electrode_manager
rcParams = {}

# which inversion code should be used to compute geometric factors
# for now this will stay at 'crtomo'
rcParams['geom_factor.inversion_code'] = 'crtomo'

# this is just to silence pep8/flake errors
reda.main.logger
EnterDirectory
CreateEnterDirectory
BaseContainer
ERT
TDIP
CR
SIP
sEIT
test
search
data
ConfigManager
transforms
eit_fzj
fzj_readbin
electrode_manager


def version():
    """Return the installed version of reda, using the pkg_resources package"""
    import pkg_resources
    return pkg_resources.require('reda')[0].version



import reda.main.units as units
import numpy as np
import pandas as pd

import reda.utils.mpl
plt, mpl = reda.utils.mpl.setup()


def _get_unique_identifiers(ee_raw):
    """

    """
    ee = ee_raw.copy()

    # sort order of dipole electrodes, i.e., turn 2-1 into 1-2
    ee_s = np.sort(ee, axis=1)

    # get unique dipoles
    eeu = np.unique(
        ee_s.view(ee_s.dtype.descr * 2)
    ).view(ee_s.dtype).reshape(-1, 2)

    # sort according to first electrode number
    eeu_s = eeu[np.argsort(eeu[:, 0]), :]

    # differences
    eeu_diff = np.abs(eeu_s[:, 0] - eeu_s[:, 1])
    # important: use mergesort here, as this is a stable sort algorithm,
    # i.e., it preserves the order of equal values
    indices = np.argsort(eeu_diff, kind='mergesort')

    # final arrangement
    eeu_final = eeu_s[indices, :]

    ee_ids = {
        key: value for key, value in zip(
            (eeu_final[:, 0] * 1e5 + eeu_final[:, 1]).astype(int),
            range(0, eeu_final.shape[0]),
        )
    }
    return ee_ids


def test_get_unique_identifiers(c):
    np.random.seed(1)
    results = []
    for i in range(0, 10):
        ab = np.random.permutation(c[:, 0:2])
        print(ab)
        q = _get_unique_identifiers(ab)
        for key in sorted(q.keys()):
            print(key, q[key])
        results.append(q)
    # compare the results
    print('checking results:')
    for x in results:
        if x != results[0]:
            print('error')


def plot_pseudosection_type2(dataobj, column, **kwargs):
    """Create a pseudosection plot of type 2.

    For a given measurement data set, create plots that graphically show
    the data in a 2D color plot. Hereby, x and y coordinates in the plot
    are determined by the current dipole (x-axis) and voltage dipole
    (y-axis) of the corresponding measurement configurations.

    This type of rawdata plot can plot any type of measurement
    configurations, i.e., it is not restricted to certain types of
    configurations such as Dipole-dipole or Wenner configurations. However,
    spatial inferences cannot be made from the plots for all configuration
    types.

    Coordinates are generated by separately sorting the dipoles
    (current/voltage) along the first electrode, and then subsequently
    sorting by the difference (skip) between both electrodes.

    Note that this type of raw data plot does not take into account the
    real electrode spacing of the measurement setup.

    Type 2 plots can show normal and reciprocal data at the same time.
    Hereby the upper left triangle of the plot area usually contains normal
    data, and the lower right triangle contains the corresponding
    reciprocal data. Therefore a quick assessment of normal-reciprocal
    differences can be made by visually comparing the symmetry on the 1:1
    line going from the lower left corner to the upper right corner.

    Note that this interpretation usually only holds for Dipole-Dipole data
    (and the related Wenner configurations).

    Parameters
    ----------

    dataobj: ERT container|pandas.DataFrame
        Container or DataFrame with data to plot
    column: string
        Column key to plot
    ax: matplotlib.Axes object, optional
        axes object to plot to. If not provided, a new figure and axes
        object will be created and returned
    nocb: bool, optional
        if set to False, don't plot the colorbar
    cblabel: string, optional
        label for the colorbar
    cbmin: float, optional
        colorbar minimum
    cbmax: float, optional
        colorbar maximum
    xlabel: string, optional
        xlabel for the plot
    ylabel: string, optional
        ylabel for the plot
    do_not_saturate: bool, optional
        if set to True, then values outside the colorbar range will not
        saturate with the respective limit colors. Instead, values lower
        than the CB are colored "cyan" and vaues above the CB limit are
        colored "red"
    log10: bool, optional
        if set to True, plot the log10 values of the provided data

    Returns
    -------
    fig:
        figure object
    ax:
        axes object
    cb:
        colorbar object

    Examples
    --------

    You can just supply a pandas.DataFrame to the plot function:

    .. plot::
        :include-source:

        import numpy as np
        configs = np.array((
            (1, 2, 4, 3),
            (1, 2, 5, 4),
            (1, 2, 6, 5),
            (2, 3, 5, 4),
            (2, 3, 6, 5),
            (3, 4, 6, 5),
        ))
        measurements = np.random.random(configs.shape[0])
        import pandas as pd
        df = pd.DataFrame(configs, columns=['a', 'b', 'm', 'n'])
        df['measurements'] = measurements

        from reda.plotters.pseudoplots import plot_pseudosection_type2
        fig, ax, cb = plot_pseudosection_type2(
           dataobj=df,
           column='measurements',
        )

    You can also supply axes to plot to:

    .. plot::
        :include-source:

        import numpy as np
        configs = np.array((
            (1, 2, 4, 3),
            (1, 2, 5, 4),
            (1, 2, 6, 5),
            (2, 3, 5, 4),
            (2, 3, 6, 5),
            (3, 4, 6, 5),
        ))
        measurements = np.random.random(configs.shape[0])
        measurements2 = np.random.random(configs.shape[0])

        import pandas as pd
        df = pd.DataFrame(configs, columns=['a', 'b', 'm', 'n'])
        df['measurements'] = measurements
        df['measurements2'] = measurements2

        from reda.plotters.pseudoplots import plot_pseudosection_type2

        fig, axes = plt.subplots(1, 2)

        plot_pseudosection_type2(
            df,
            column='measurements',
            ax=axes[0],
            cblabel='this label',
            xlabel='xlabel',
            ylabel='ylabel',
        )
        plot_pseudosection_type2(
            df,
            column='measurements2',
            ax=axes[1],
            cblabel='measurement 2',
            xlabel='xlabel',
            ylabel='ylabel',
        )
        fig.tight_layout()

    >>> from reda.testing.containers import ERTContainer_nr
    >>> import reda.plotters.pseudoplots as ps
    >>> fig, axes, cb = ps.plot_pseudosection_type2(ERTContainer_nr, 'r')

    """
    if isinstance(dataobj, pd.DataFrame):
        df = dataobj
    else:
        df = dataobj.data

    c = df[['a', 'b', 'm', 'n']].values

    AB_ids = _get_unique_identifiers(c[:, 0:2])
    MN_ids = _get_unique_identifiers(c[:, 2:4])

    ab_sorted = np.sort(c[:, 0:2], axis=1)
    mn_sorted = np.sort(c[:, 2:4], axis=1)

    AB_coords = [
        AB_ids[x] for x in
        (ab_sorted[:, 0] * 1e5 + ab_sorted[:, 1]).astype(int)
    ]
    MN_coords = [
        MN_ids[x] for x in
        (mn_sorted[:, 0] * 1e5 + mn_sorted[:, 1]).astype(int)
    ]

    # check for duplicate positions
    ABMN_coords = np.vstack((AB_coords, MN_coords)).T.copy()
    _, counts = np.unique(
        ABMN_coords.view(
            ABMN_coords.dtype.descr * 2
        ),
        return_counts=True,
    )
    # import IPython
    # IPython.embed()
    # exit()
    if np.any(counts > 1):
        print('found duplicate coordinates!')
        # duplicate_configs = np.where(counts > 1)[0]
        # print('duplicate configs:')
        # print('A B M N')
        # for i in duplicate_configs:
        #     print(c[i, :])

    # prepare matrix
    plot_values = np.squeeze(df[column].values)

    if kwargs.get('log10', False):
        plot_values = np.log10(plot_values)

    C = np.zeros((len(MN_ids.items()), len(AB_ids))) * np.nan
    C[MN_coords, AB_coords] = plot_values

    # for display purposes, reverse the first dimension
    C = C[::-1, :]

    ax = kwargs.get('ax', None)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(15 / 2.54, 10 / 2.54))
    fig = ax.get_figure()

    cmap = mpl.cm.get_cmap('viridis')
    if kwargs.get('do_not_saturate', False):
        cmap.set_over(
            color='r'
        )
        cmap.set_under(
            color='c'
        )
    im = ax.matshow(
        C,
        interpolation='none',
        cmap=cmap,
        aspect='auto',
        vmin=kwargs.get('cbmin', None),
        vmax=kwargs.get('cbmax', None),
        extent=[
            0, max(AB_coords),
            0, max(MN_coords),
        ],
    )

    max_xy = max((max(AB_coords), max(MN_coords)))
    ax.plot(
        (0, max_xy),
        (0, max_xy),
        '-',
        color='k',
        linewidth=1.0,
    )

    cb = None
    if not kwargs.get('nocb', False):
        cb = fig.colorbar(im, ax=ax)
        label = units.get_label(column)
        if not mpl.rcParams['text.usetex']:
            label = label.replace('_', '-')
        cb.set_label(
            kwargs.get('cblabel', label)
        )

    ax.set_xlabel(
        kwargs.get('xlabel', 'current dipoles')
    )
    ax.set_ylabel(
        kwargs.get('ylabel', 'voltage dipoles')
    )

    return fig, ax, cb


def plot_pseudosection_type3(dataobj, column, **kwargs):
    """Create a pseudosection plot of type 3.

    For a given measurement data set, create plots that graphically show
    the data in a 2D pseudoplot. Hereby, x and y coordinates (pseudodistance
    and pseudodepth) in the plot are determined by the corresponding measurement
    configuration (after Roy and Apparao (1971) and Dahlin and Zou (2005)).

    This type of rawdata plot can plot any type of measurement
    configurations, i.e., it is not restricted to certain types of
    configurations such as Dipole-dipole or Wenner configurations.

    Parameters
    ----------

    dataobj: ERT container|pandas.DataFrame
        Container or DataFrame with data to plot
    column: string
        Column key to plot
    ax: matplotlib.Axes object, optional
        axes object to plot to. If not provided, a new figure and axes
        object will be created and returned
    nocb: bool, optional
        if set to False, don't plot the colorbar
    cblabel: string, optional
        label for the colorbar
    cbmin: float, optional
        colorbar minimum
    cbmax: float, optional
        colorbar maximum
    xlabel: string, optional
        xlabel for the plot
    ylabel: string, optional
        ylabel for the plot
    do_not_saturate: bool, optional
        if set to True, then values outside the colorbar range will not
        saturate with the respective limit colors. Instead, values lower
        than the CB are colored "cyan" and vaues above the CB limit are
        colored "red"
    markersize: float, optional
        size of plotted data points
    spacing: float, optional
        if set to True, the actual electrode spacing is used for the computation
        of the pseudodepth and -distance; default value is 1 m
    log10: bool, optional
        if set to True, plot the log10 values of the provided data

    Returns
    -------
    fig:
        figure object
    ax:
        axes object
    cb:
        colorbar object

    Examples
    --------

    You can just supply a pandas.DataFrame to the plot function:

    .. plot::
        :include-source:

        import numpy as np
        configs = np.array((
            (1, 2, 4, 3),
            (1, 2, 5, 4),
            (1, 2, 6, 5),
            (2, 3, 5, 4),
            (2, 3, 6, 5),
            (3, 4, 6, 5),
        ))
        measurements = np.random.random(configs.shape[0])
        import pandas as pd
        df = pd.DataFrame(configs, columns=['a', 'b', 'm', 'n'])
        df['measurements'] = measurements

        from reda.plotters.pseudoplots import plot_pseudosection_type3
        fig, ax, cb = plot_pseudosection_type3(
           dataobj=df,
           column='measurements',
        )

    You can also supply axes to plot to:

    .. plot::
        :include-source:

        import numpy as np
        configs = np.array((
            (1, 2, 4, 3),
            (1, 2, 5, 4),
            (1, 2, 6, 5),
            (2, 3, 5, 4),
            (2, 3, 6, 5),
            (3, 4, 6, 5),
        ))
        measurements = np.random.random(configs.shape[0])
        measurements2 = np.random.random(configs.shape[0])

        import pandas as pd
        df = pd.DataFrame(configs, columns=['a', 'b', 'm', 'n'])
        df['measurements'] = measurements
        df['measurements2'] = measurements2

        from reda.plotters.pseudoplots import plot_pseudosection_type3

        fig, axes = plt.subplots(1, 2)

        plot_pseudosection_type3(
            df,
            column='measurements',
            ax=axes[0],
            cblabel='this label',
            xlabel='xlabel',
            ylabel='ylabel',
        )
        plot_pseudosection_type3(
            df,
            column='measurements2',
            ax=axes[1],
            cblabel='measurement 2',
            xlabel='xlabel',
            ylabel='ylabel',
        )
        fig.tight_layout()
    """
    if isinstance(dataobj, pd.DataFrame):
        df = dataobj
    else:
        df = dataobj.data

    c = (df[['a', 'b', 'm', 'n']]-1)*kwargs.get('spacing', 1)

    # define the configuration
    # check on first quadrupole and assume the config is consistent
    # dipole-dipole
    if (sum(np.greater([c.a[0], c.b[0]], [c.m[0], c.n[0]])) == 2 or
            sum(np.greater([c.m[0], c.n[0]], [c.a[0], c.b[0]])) == 2):
        c.loc[:, 'xp'] = ((c.a + c.b)/2 + (c.n + c.m)/2) / 2
        c.loc[:, 'zp'] = -((c.n - c.b)*0.195)  # Roi and Appparo
    # multiple gradient
    else:
        xmn = (c.m + c.n) / 2
        c.loc[:, 'xp'] = xmn
        c.loc[:, 'zp'] = np.abs(np.min([xmn-c.a, c.b-xmn], axis=0) / 3)*-1  # Dahlin, Zhou

    # extract the values to plot
    c['plot_values'] = df[column]

    if kwargs.get('log10', False):
        c['plot_values'] = np.log10(c['plot_values'])

    # sort after the pseudodistance and pseudodepth
    pseudocoords = c[['xp', 'zp', 'plot_values']].sort_values(by=['xp', 'zp'])

    ax = kwargs.get('ax', None)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(15 / 2.54, 10 / 2.54))
    fig = ax.get_figure()

    cmap = mpl.cm.get_cmap('viridis')
    if kwargs.get('do_not_saturate', False):
        cmap.set_over(
            color='r'
        )
        cmap.set_under(
            color='c'
        )

    pseudocoords['markersize'] = kwargs.get('markersize', 10)
    # check for same pseudocoordinates
    common_pscoords = pseudocoords[['xp', 'zp']].drop_duplicates()

    def smallify(markersize_array):
        """
        For a given array of markersizes (with same size) compute a stepwise
        decreasing factor for the plotting size based on the length of the array.
        """
        nelems = len(markersize_array)
        factor = 1/nelems
        return markersize_array*np.flip(np.arange(1, nelems+1)*factor)

    # check for overlapping points and adjust markersize accordingly
    for _, common in common_pscoords.iterrows():
        subset = pseudocoords.query('xp == {} and zp == {}'.format(common.xp, common.zp))
        pseudocoords.loc[subset.index, 'markersize'] = smallify(
            pseudocoords.loc[subset.index, 'markersize'].values)

    scat = ax.scatter(pseudocoords['xp'], pseudocoords['zp'],
                      s=pseudocoords['markersize'],
                      c=pseudocoords['plot_values'],
                      marker='o',
                      edgecolors='none',
                      cmap=cmap,
                      vmin=kwargs.get('cbmin', None),
                      vmax=kwargs.get('cbmax', None),
                      alpha=1)

    ax.set_xlim([np.min(c['xp'])-1*kwargs.get('spacing', 1),
                 np.max(c['xp'])+1*kwargs.get('spacing', 1)])
    ax.set_ylim([np.min(c['zp'])-1*kwargs.get('spacing', 1),
                 np.max(c['zp'])+1*kwargs.get('spacing', 1)])

    cb = None
    if not kwargs.get('nocb', False):
        cb = fig.colorbar(scat, ax=ax)
        cb.set_label(
            kwargs.get('cblabel', units.get_label(column))
        )

    ax.set_xlabel(
        kwargs.get('xlabel', 'Pseudodistance')
    )
    ax.set_ylabel(
        kwargs.get('ylabel', 'Pseudodepth')
    )

    return fig, ax, cb


def plot_ps_extra(dataobj, key, **kwargs):
    """Create grouped pseudoplots for one or more time steps

    Parameters
    ----------
    dataobj: :class:`reda.containers.ERT`
        An ERT container with loaded data
    key: string
        The column name to plot
    subquery: string, optional
    cbmin: float, optional
    cbmax: float, optional

    Examples
    --------
    >>> import reda.testing.containers
    >>> ert = reda.testing.containers.ERTContainer_nr
    >>> import reda.plotters.pseudoplots as PS
    >>> fig = PS.plot_ps_extra(ert, key='r')
    """
    if isinstance(dataobj, pd.DataFrame):
        df_raw = dataobj
    else:
        df_raw = dataobj.data

    if kwargs.get('subquery', False):
        df = df_raw.query(kwargs.get('subquery'))
    else:
        df = df_raw

    def fancyfy(axes, N):
        for ax in axes[0:-1, :].flat:
            ax.set_xlabel('')
        for ax in axes[:, 1:].flat:
            ax.set_ylabel('')

    g = df.groupby('timestep')
    N = len(g.groups.keys())
    nrx = min((N, 5))
    nry = int(np.ceil(N / nrx))
    # the sizes are heuristics [inches]
    sizex = nrx * 3
    sizey = nry * 4 - 1
    fig, axes = plt.subplots(
        nry, nrx,
        sharex=True,
        sharey=True,
        figsize=(sizex, sizey),
    )
    axes = np.atleast_2d(axes)

    cbs = []
    for ax, (name, group) in zip(axes.flat, g):
        fig1, axes1, cb1 = plot_pseudosection_type2(
            group,
            key,
            ax=ax,
            log10=False,
            cbmin=kwargs.get('cbmin', None),
            cbmax=kwargs.get('cbmax', None),
        )
        cbs.append(cb1)
        ax.set_title('timestep: {0}'.format(int(name)))
        ax.xaxis.set_ticks_position('bottom')
        ax.set_aspect('equal')

    for cb in np.array(cbs).reshape(axes.shape)[:, 0:-1].flat:
        cb.ax.set_visible(False)

    fancyfy(axes, N)
    fig.tight_layout()
    return fig



# *-* coding: utf-8 *-*
"""Manage measurement configurations measurements.
"""
import itertools

import numpy as np
import pandas as pd

import reda.utils.mpl
from reda.utils import opt_import

plt, mpl = reda.utils.mpl.setup()


class ConfigManager(object):
    """The class`ConfigManager` manages four-point measurement configurations.
    """

    def __init__(self, nr_of_electrodes=None):
        # store the configs as a Nx4 numpy array
        self.configs = None
        # each measurement can store additional data here
        self.metadata = {}
        # number of electrodes
        self.nr_electrodes = nr_of_electrodes

    def abmn_to_dataframe(self):
        """Return a pandas.DataFrame containing the measurement configurations

        Returns
        -------
        abmn_df : pandas.DataFrame
            Configurations in a DataFrame (columns: a,b,m,n)
        """
        abmn_df = pd.DataFrame(self.configs, columns=['a', 'b', 'm', 'n'])
        return abmn_df

    def _get_next_index(self):
        """
        """
        self.meas_counter += 1
        return self.meas_counter

    def clear_configs(self):
        """Remove all configs. This implies deleting all measurements.
        """
        del (self.configs)
        self.configs = None

    @property
    def nr_of_configs(self):
        """Return number of configurations

        Returns
        -------
        nr_of_configs: int
            number of configurations stored in this instance

        """
        if self.configs is None:
            return 0
        else:
            return self.configs.shape[0]

    def _crmod_to_abmn(self, configs):
        r"""convert crmod-style configurations to a Nx4 array

        CRMod-style configurations merge A and B, and M and N, electrode
        numbers into one large integer each:

        .. math ::

            AB = A \cdot 10^4 + B

            MN = M \cdot 10^4 + N

        Parameters
        ----------
        configs: numpy.ndarray
            Nx2 array holding the configurations to convert

        Examples
        --------

        >>> import numpy as np
        >>> from reda.configs.configManager import ConfigManager
        >>> config = ConfigManager(nr_of_electrodes=5)
        >>> crmod_configs = np.array((
        ...     (10002, 40003),
        ...     (10010, 30004),
        ... ))
        >>> abmn = config._crmod_to_abmn(crmod_configs)
        >>> print(abmn)   # doctest: +NORMALIZE_WHITESPACE
        [[ 1  2  4  3]
         [ 1 10  3  4]]

        """
        A = np.floor(configs[:, 0] / 1e4).astype(int)
        B = (configs[:, 0] % 1e4).astype(int)
        M = np.floor(configs[:, 1] / 1e4).astype(int)
        N = (configs[:, 1] % 1e4).astype(int)
        ABMN = np.hstack((
            A[:, np.newaxis],
            B[:, np.newaxis],
            M[:, np.newaxis],
            N[:, np.newaxis]
        )).astype(int)
        return ABMN

    def load_configs(self, filename):
        """Load configurations from a file with four columns: a b m n
        """
        configs = np.loadtxt(filename)
        self.add_to_configs(configs)

    def load_injections_from_mcf(self, filename):
        """Load injections from a mcf-file used by the sEIT-systems from FZJ;
        Injections in the file must be formatted like the following:
        SE 001 002
        SE 002 001
        .
        .
        .

        Parameters
        ----------
        filename: string
            absolute or relative path to a mcf-file

        Returns
        -------
        Nx2 numpy array with current injections
        """

        injections = []
        # load all injections
        with open(filename, encoding="latin-1") as mcf:
            for line in mcf:
                if line[:2] == "SE":
                    injections.append((line[3:6], line[7:10]))

        # convert to array
        injections = np.asarray(injections, dtype=int)
        # remove double entries because only one current injection
        # is needed for configs
        mask = np.less(injections[:, 0], injections[:, 1])
        injections = injections[mask]

        return(injections)

    def load_crmod_config(self, filename):
        """Load a CRMod configuration file

        Parameters
        ----------
        filename: string
            absolute or relative path to a crmod config.dat file

        """
        with open(filename, 'r') as fid:
            nr_of_configs = int(fid.readline().strip())
            configs = np.loadtxt(fid)
            print('loaded configs:', configs.shape)
            if nr_of_configs != configs.shape[0]:
                raise Exception(
                    'indicated number of measurements does not equal ' +
                    'to actual number of measurements')
            ABMN = self._crmod_to_abmn(configs[:, 0:2])
            self.configs = ABMN

    def load_crmod_or_4c_configs(self, filename):
        """Load configurations either from a CRMod measurement file, or from a
        four-column file. Assume 1-indexed data (start with electrode 1).

        Parameters
        ----------
        filename : str
            Path to file that shall be imported

        Returns
        -------
        None
        """
        # The first line decides
        first_line = np.genfromtxt(filename, max_rows=1)
        if first_line.size == 4:
            self.load_configs(filename)
        elif first_line.size == 1:
            self.load_crmod_config(filename)
        else:
            raise Exception('File format not recognized')

    # def load_crmod_volt(self, filename):
    #     """Load a CRMod measurement file (commonly called volt.dat)

    #     Parameters
    #     ----------
    #     filename: string
    #         path to filename

    #     Returns
    #     -------
    #     list
    #         list of measurement ids
    #     """
    #     with open(filename, 'r') as fid:
    #         nr_of_configs = int(fid.readline().strip())
    #         measurements = np.loadtxt(fid)
    #         if nr_of_configs != measurements.shape[0]:
    #             raise Exception(
    #                 'indicated number of measurements does not equal ' +
    #                 'to actual number of measurements')
    #     ABMN = self._crmod_to_abmn(measurements[:, 0:2])
    #     if self.configs is None:
    #         self.configs = ABMN
    #     else:
    #         # check that configs match
    #         if not np.all(ABMN == self.configs):
    #             raise Exception(
    #                 'previously stored configurations do not match new ' +
    #                 'configurations')

    #     # add data
    #     cid_mag = self.add_measurements(measurements[:, 2])
    #     cid_pha = self.add_measurements(measurements[:, 3])
    #     return [cid_mag, cid_pha]

    def _get_crmod_abmn(self):
        """return a Nx2 array with the measurement configurations formatted
        CRTomo style
        """
        ABMN = np.vstack((
            self.configs[:, 0] * 1e4 + self.configs[:, 1],
            self.configs[:, 2] * 1e4 + self.configs[:, 3],
        )).T.astype(int)
        return ABMN

    def write_crmod_volt(self, filename, mid):
        """Write the measurements to the output file in the volt.dat file
        format that can be read by CRTomo.

        Parameters
        ----------
        filename: string
            output filename
        mid: int or [int, int]
            measurement ids of magnitude and phase measurements. If only one ID
            is given, then the phase column is filled with zeros

        """
        ABMN = self._get_crmod_abmn()

        if isinstance(mid, (list, tuple)):
            mag_data = self.measurements[mid[0]]
            pha_data = self.measurements[mid[1]]
        else:
            mag_data = self.measurements[mid]
            pha_data = np.zeros(mag_data.shape)

        all_data = np.hstack((ABMN, mag_data[:, np.newaxis],
                              pha_data[:, np.newaxis]))

        with open(filename, 'wb') as fid:
            fid.write(bytes(
                '{0}\n'.format(ABMN.shape[0]),
                'utf-8',
            ))
            np.savetxt(fid, all_data, fmt='%i %i %f %f')

    def write_crmod_volt_with_individual_errors(
            self, filename, data_mids, error_mids, norm_mag=1, norm_pha=1,
            ):
        """

        """
        ABMN = self._get_crmod_abmn()

        if isinstance(data_mids, (list, tuple)):
            mag_data = self.measurements[data_mids[0]]
            pha_data = self.measurements[data_mids[1]]
            mag_error = self.measurements[error_mids[0]]
            pha_error = self.measurements[error_mids[1]]
        else:
            raise Exception(
                'Individual errors for mag-only not implemented yet')
            mag_data = self.measurements[data_mids]
            pha_data = np.zeros(data_mids.shape)

        all_data = np.hstack(
            (
                ABMN,
                mag_data[:, np.newaxis],
                pha_data[:, np.newaxis],
                mag_error[:, np.newaxis],
                pha_error[:, np.newaxis],
            )
        )

        with open(filename, 'wb') as fid:
            fid.write(bytes(
                '{0} T\n'.format(ABMN.shape[0]),
                'utf-8',
            ))
            np.savetxt(fid, all_data, fmt='%i %i %f %f %f %f')
            fid.write(
                bytes(
                    '{} {}\n'.format(norm_mag, norm_pha),
                    'utf-8',
                )
            )

    def write_crmod_config(self, filename):
        """Write the configurations to a configuration file in the CRMod format
        All configurations are merged into one previor to writing to file

        Parameters
        ----------
        filename: string
            absolute or relative path to output filename (usually config.dat)
        """
        ABMN = self._get_crmod_abmn()

        with open(filename, 'wb') as fid:
            fid.write(bytes(
                '{0}\n'.format(ABMN.shape[0]),
                'utf-8',
            ))
            np.savetxt(fid, ABMN.astype(int), fmt='%i %i')

    def gen_dipole_dipole(self,
                          skipc,
                          skipv=None,
                          stepc=1,
                          stepv=1,
                          nr_voltage_dipoles=10,
                          before_current=False,
                          allow_crossings=False,
                          start_skip=0,
                          N=None):
        """Generate dipole-dipole configurations

        Parameters
        ----------
        skipc: int
            number of electrode positions that are skipped between electrodes
            of a given dipole
        skipv: int
            steplength between subsequent voltage dipoles. A steplength of 0
            will produce increments by one, i.e., 3-4, 4-5, 5-6 ...
        stepc: int
            steplength between subsequent current dipoles. A steplength of 0
            will produce increments by one, i.e., 3-4, 4-5, 5-6 ...
        stepv: int
            steplength between subsequent voltage dipoles. A steplength of 0
            will produce increments by one, i.e., 3-4, 4-5, 5-6 ...
        nr_voltage_dipoles: int
            the number of voltage dipoles to generate for each current
            injection dipole
        before_current: bool, optional
            if set to True, also generate voltage dipoles in front of current
            dipoles.
        allow_crossings : bool, optional
            Allow potential dipoles to cross current injection dipoles. Note
            that any quadpole generated by crossing the current electrodes is
            NOT counted into nr_voltage_dipoles.
        start_skip: int, optional
            how many electrode to skip before/after the first/second current
            electrode.
        N: int, optional
            number of electrodes, must be given if not already known by the
            config instance

        Examples
        --------

        >>> from reda.configs.configManager import ConfigManager
        >>> config = ConfigManager(nr_of_electrodes=10)
        >>> config.gen_dipole_dipole(skipc=2)
        array([[ 1,  4,  5,  8],
               [ 1,  4,  6,  9],
               [ 1,  4,  7, 10],
               [ 2,  5,  6,  9],
               [ 2,  5,  7, 10],
               [ 3,  6,  7, 10]])

        >>> from reda.configs.configManager import ConfigManager
        >>> config = ConfigManager(nr_of_electrodes=10)
        >>> config.gen_dipole_dipole(
        ...     skipc=2, before_current=True, allow_crossings=True)
        array([[ 1,  4,  2,  5],
               [ 1,  4,  3,  6],
               [ 1,  4,  5,  8],
               [ 1,  4,  6,  9],
               [ 1,  4,  7, 10],
               [ 2,  5,  1,  4],
               [ 2,  5,  3,  6],
               [ 2,  5,  4,  7],
               [ 2,  5,  6,  9],
               [ 2,  5,  7, 10],
               [ 3,  6,  1,  4],
               [ 3,  6,  2,  5],
               [ 3,  6,  4,  7],
               [ 3,  6,  5,  8],
               [ 3,  6,  7, 10]])

        >>> from reda.configs.configManager import ConfigManager
        >>> config = ConfigManager(nr_of_electrodes=10)
        >>> config.gen_dipole_dipole(skipc=1, before_current=True)
        array([[ 1,  3,  4,  6],
               [ 1,  3,  5,  7],
               [ 1,  3,  6,  8],
               [ 1,  3,  7,  9],
               [ 1,  3,  8, 10],
               [ 2,  4,  5,  7],
               [ 2,  4,  6,  8],
               [ 2,  4,  7,  9],
               [ 2,  4,  8, 10],
               [ 3,  5,  6,  8],
               [ 3,  5,  7,  9],
               [ 3,  5,  8, 10],
               [ 4,  6,  1,  3],
               [ 4,  6,  7,  9],
               [ 4,  6,  8, 10],
               [ 5,  7,  2,  4],
               [ 5,  7,  1,  3],
               [ 5,  7,  8, 10]])

        """
        if N is None and self.nr_electrodes is None:
            raise Exception('You must provide the number of electrodes')
        elif N is None:
            N = self.nr_electrodes

        # by default, current voltage dipoles have the same size
        if skipv is None:
            skipv = skipc

        configs = []
        # current dipoles
        for a in range(0, N - skipv - skipc - 3, stepc):
            b = a + skipc + 1
            nr = 0
            # potential dipoles before current injection
            if before_current:
                for n in range(a - start_skip - 1, -1, -stepv):
                    nr += 1
                    if nr > nr_voltage_dipoles:
                        continue
                    m = n - skipv - 1
                    if m < 0:
                        continue
                    quadpole = np.array((a, b, m, n)) + 1
                    configs.append(quadpole)
                if allow_crossings:
                    # backward crossings
                    electrodes_between_injections = list(range(a + 1, b))
                    for n in electrodes_between_injections:
                        m = n - skipv - 1
                        if m < 0:
                            continue
                        quadpole = np.array((a, b, m, n)) + 1
                        configs.append(quadpole)

            if allow_crossings:
                # forward crossings
                electrodes_between_injections = list(range(a + 1, b))
                for m in electrodes_between_injections:
                    n = m + skipv + 1
                    quadpole = np.array((a, b, m, n)) + 1
                    configs.append(quadpole)

            # potential dipoles after current injection
            nr = 0
            for m in range(b + start_skip + 1, N - skipv - 1, stepv):
                nr += 1
                if nr > nr_voltage_dipoles:
                    continue
                n = m + skipv + 1
                quadpole = np.array((a, b, m, n)) + 1
                configs.append(quadpole)

        configs = np.array(configs)
        # now add to the instance
        if self.configs is None:
            self.configs = configs
        else:
            self.configs = np.vstack((self.configs, configs))
        return configs

    def gen_inner_gradients(self, injections, skips=None):
        """
        Given a set of current injections, generate voltage dipoles between the
        current electrodes

        Examples
        --------
        >>> import reda
        >>> cfg = reda.ConfigManager(nr_of_electrodes=10)
        >>> cfg.gen_inner_gradients([1, 10], skips=[2, ])
        array([[ 1, 10,  2,  5],
               [ 1, 10,  3,  6],
               [ 1, 10,  4,  7],
               [ 1, 10,  5,  8],
               [ 1, 10,  6,  9]])


        Parameters
        ----------
        injections : Nx2 numpy.ndarray
            Current injections
        skips : None or iterable
            The skips to generate. None will generate all possible skips

        Returns
        -------
        configs : Nx4 numpy.ndarray
            The generated configurations

        """
        configs_raw = []
        for ab in np.atleast_2d(injections):
            a = min(ab)
            b = max(ab)
            if b - a < 3:
                continue

            max_skip = b - a - 2
            if skips is None:
                use_skips = range(0, max_skip)
            else:
                use_skips = [x for x in skips if x <= max_skip]
            for skip in use_skips:
                for m in range(a + 1, b - 1 - skip, 1):
                    n = m + skip + 1
                    configs_raw.append(
                        (ab[0], ab[1], m, n)
                    )
        configs = np.array(configs_raw)
        self.add_to_configs(configs)
        return configs

    def gen_gradient(self, skip=0, step=1, vskip=0, vstep=1):
        """Generate gradient measurements

        Parameters
        ----------
        skip: int
            distance between current electrodes
        step: int
            steplength between subsequent current dipoles
        vskip: int
            distance between voltage electrodes
        vstep: int
            steplength between subsequent voltage dipoles

        """
        N = self.nr_electrodes
        quadpoles = []
        for a in range(1, N - skip, step):
            b = a + skip + 1
            for m in range(a + 1, b - vskip - 1, vstep):
                n = m + vskip + 1
                quadpoles.append((a, b, m, n))

        configs = np.array(quadpoles)
        if configs.size == 0:
            return None

        self.add_to_configs(configs)
        return configs

    def gen_all_voltages_for_injections(self, injections_raw):
        r"""For a given set of current injections AB, generate all possible
        unique potential measurements.

        After Noel and Xu, 1991, for N electrodes, the number of possible
        voltage dipoles for a given current dipole is :math:`(N - 2)(N - 3) /
        2`. This includes normal and reciprocal measurements.

        If current dipoles are generated with
        ConfigManager.gen_all_current_dipoles(), then :math:`N \cdot (N - 1) /
        2` current dipoles are generated. Thus, this function will produce
        :math:`(N - 1)(N - 2)(N - 3) / 4` four-point configurations ABMN, half
        of which are reciprocals (Noel and Xu, 1991).

        All generated measurements are added to the instance.

        Use ConfigManager.split_into_normal_and_reciprocal() to split the
        configurations into normal and reciprocal measurements.

        Parameters
        ----------
        injections : numpy.ndarray
            Kx2 array holding K current injection dipoles A-B

        Returns
        -------
        configs : numpy.ndarray
            Nax4 array holding all possible measurement configurations

        """
        if isinstance(injections_raw, (list, tuple)):
            injections = np.atleast_2d(np.array(injections_raw)).astype(int)
        else:
            injections = np.atleast_2d(injections_raw).astype(int)

        N = self.nr_electrodes
        all_quadpoles = []
        for idipole in injections:
            # sort current electrodes and convert to array indices
            Ipoles = np.sort(idipole) - 1

            # voltage electrodes
            velecs = list(range(1, N + 1))

            # remove current electrodes
            del (velecs[Ipoles[1]])
            del (velecs[Ipoles[0]])

            # permutate remaining
            voltages = itertools.permutations(velecs, 2)
            for voltage in voltages:
                all_quadpoles.append((idipole[0], idipole[1], voltage[0],
                                      voltage[1]))
        configs_unsorted = np.array(all_quadpoles)
        # sort AB and MN
        configs_sorted = np.hstack((
            np.sort(configs_unsorted[:, 0:2], axis=1),
            np.sort(configs_unsorted[:, 2:4], axis=1),
        ))
        configs = self.remove_duplicates(configs_sorted)

        self.add_to_configs(configs)
        self.remove_duplicates()
        return configs

    def gen_all_current_dipoles(self):
        r"""Generate all possible current dipoles for the given number of
        electrodes (self.nr_electrodes). Duplicates are removed in the process.

        After Noel and Xu, 1991, for N electrodes, the number of possible
        unique configurations is :math:`N \cdot (N - 1) / 2`. This excludes
        duplicates in the form of switches current/voltages electrodes, as well
        as reciprocal measurements.

        Returns
        -------
        configs : Nx2 numpy.ndarray
            all possible current dipoles A-B
        """
        N = self.nr_electrodes
        celecs = list(range(1, N + 1))
        AB_list = itertools.permutations(celecs, 2)
        AB = np.array([ab for ab in AB_list])
        AB.sort(axis=1)

        # now we need to filter duplicates
        AB = np.unique(AB.view(AB.dtype.descr * 2)).view(AB.dtype).reshape(
            -1, 2)

        return AB

    def remove_duplicates(self, configs=None):
        """remove duplicate entries from 4-point configurations. If no
        configurations are provided, then use self.configs. Unique
        configurations are only returned if configs is not None.

        Parameters
        ----------
        configs: Nx4 numpy.ndarray, optional
            remove duplicates from these configurations instead from
            self.configs.

        Returns
        -------
        configs_unique: Kx4 numpy.ndarray
            unique configurations. Only returned if configs is not None

        """
        if configs is None:
            c = self.configs
        else:
            c = configs
        struct = c.view(c.dtype.descr * 4)
        configs_unique = np.unique(struct).view(c.dtype).reshape(-1, 4)
        if configs is None:
            self.configs = configs_unique
        else:
            return configs_unique

    def gen_schlumberger(self, M, N, a=None):
        """generate one Schlumberger sounding configuration, that is, one set
        of configurations for one potential dipole MN.

        Parameters
        ----------
        M: int
            electrode number for the first potential electrode
        N: int
            electrode number for the second potential electrode
        a: int, optional
            stepping between subsequent voltage electrodes. If not set,
            determine it as a = abs(M - N)

        Returns
        -------
        configs: Kx4 numpy.ndarray
            array holding the configurations

        Examples
        --------

            import reda.configs.configManager as CRconfig
            config = CRconfig.ConfigManager(nr_of_electrodes=40)
            config.gen_schlumberger(M=20, N=21)

        """
        if a is None:
            a = np.abs(M - N)

        nr_of_steps_left = int(min(M, N) - 1 / a)
        nr_of_steps_right = int((self.nr_electrodes - max(M, N)) / a)
        configs = []
        for i in range(0, min(nr_of_steps_left, nr_of_steps_right)):
            A = min(M, N) - (i + 1) * a
            B = max(M, N) + (i + 1) * a
            configs.append((A, B, M, N))
        configs = np.array(configs)
        self.add_to_configs(configs)
        return configs

    def gen_wenner(self, a):
        """Generate Wenner measurement configurations.

        Parameters
        ----------
        a: int
            distance (in electrodes) between subsequent electrodes of each
            four-point configuration.

        Returns
        -------
        configs: Kx4 numpy.ndarray
            array holding the configurations
        """
        configs = []
        for i in range(1, self.nr_electrodes - 3 * a + 1):
            configs.append((i, i + 3 * a, i + 1 * a, i + 2 * a), )
        configs = np.array(configs)
        self.add_to_configs(configs)
        return configs

    def add_to_configs(self, configs):
        """Add one or more measurement configurations to the stored
        configurations

        Parameters
        ----------
        configs : list or numpy.ndarray
            list or array of configurations

        Returns
        -------
        configs : Kx4 numpy.ndarray
            array holding all configurations of this instance
        """
        if len(configs) == 0:
            return None

        if self.configs is None:
            self.configs = np.atleast_2d(configs)
        else:
            configs = np.atleast_2d(configs)
            self.configs = np.vstack((self.configs, configs))

        # make sure we store the configs as ints
        self.configs = self.configs.astype(int)

        return self.configs

    def split_into_normal_and_reciprocal(self, pad=False,
                                         return_indices=False):
        """Split the stored configurations into normal and reciprocal
        measurements

        ** *Rule 1: the normal configuration contains the smallest electrode
        number of the four involved electrodes in the current dipole* **

        Parameters
        ----------
        pad: bool, optional
            if True, add numpy.nan values to the reciprocals for non-existent
            measuremnts
        return_indices: bool, optional
            if True, also return the indices of normal and reciprocal
            measurments. This can be used to extract corresponding
            measurements.

        Returns
        -------
        normal: numpy.ndarray
            Nnx4 array. If pad is True, then Nn == N (total number of
            unique measurements). Otherwise Nn is the number of normal
            measurements.
        reciprocal: numpy.ndarray
            Nrx4 array. If pad is True, then Nr == N (total number of
            unique measurements). Otherwise Nr is the number of reciprocal
            measurements.
        nor_indices: numpy.ndarray, optional
            Nnx1 array containing the indices of normal measurements. Only
            returned if return_indices is True.
        rec_indices: numpy.ndarray, optional
            Nrx1 array containing the indices of normal measurements. Only
            returned if return_indices is True.

        """
        # for simplicity, we create an array where AB and MN are sorted
        configs = np.hstack((np.sort(self.configs[:, 0:2], axis=1),
                             np.sort(self.configs[:, 2:4], axis=1)))

        ab_min = configs[:, 0]
        mn_min = configs[:, 2]

        # rule 1
        indices_normal = np.where(ab_min < mn_min)[0]

        # now look for reciprocals
        indices_used = []
        normal = []
        normal_indices = []
        reciprocal_indices = []
        reciprocal = []
        duplicates = []
        for index in indices_normal:
            indices_used.append(index)
            normal.append(self.configs[index, :])
            normal_indices.append(index)

            # look for reciprocal configuration
            index_rec = np.where(
                # A == M, B == N, M == A, N == B
                (configs[:, 0] == configs[index, 2]) &
                (configs[:, 1] == configs[index, 3]) &
                (configs[:, 2] == configs[index, 0]) &
                (configs[:, 3] == configs[index, 1]))[0]
            if len(index_rec) == 0 and pad:
                reciprocal.append(np.ones(4) * np.nan)
            elif len(index_rec) == 1:
                reciprocal.append(self.configs[index_rec[0], :])
                indices_used.append(index_rec[0])
                reciprocal_indices.append(index_rec[0])
            elif len(index_rec > 1):
                # take the first one
                reciprocal.append(self.configs[index_rec[0], :])
                reciprocal_indices.append(index_rec[0])
                duplicates += list(index_rec[1:])
                indices_used += list(index_rec)

        # now determine all reciprocal-only parameters
        set_all_indices = set(list(range(0, configs.shape[0])))
        set_used_indices = set(indices_used)
        reciprocal_only_indices = set_all_indices - set_used_indices
        for index in reciprocal_only_indices:
            if pad:
                normal.append(np.ones(4) * np.nan)
            reciprocal.append(self.configs[index, :])

        normals = np.array(normal)
        reciprocals = np.array(reciprocal)

        if return_indices:
            return normals, reciprocals, normal_indices, reciprocal_indices
        else:
            return normals, reciprocals

    def gen_reciprocals(self, append=False):
        """ Generate reciprocal configurations, sort by AB, and optionally
        append to configurations.

        Parameters
        ----------
        append : bool
            Append reciprocals to configs (the default is False).

        Examples
        --------
        >>> cfgs = ConfigManager(nr_of_electrodes=5)
        >>> nor = cfgs.gen_dipole_dipole(skipc=0)
        >>> rec = cfgs.gen_reciprocals(append=True)
        >>> print(cfgs.configs)
        [[1 2 3 4]
         [1 2 4 5]
         [2 3 4 5]
         [3 4 1 2]
         [4 5 1 2]
         [4 5 2 3]]
        """
        # Switch AB and MN
        reciprocals = self.configs.copy()[:, ::-1]
        reciprocals[:, 0:2] = np.sort(reciprocals[:, 0:2], axis=1)
        reciprocals[:, 2:4] = np.sort(reciprocals[:, 2:4], axis=1)
        # # Sort by current dipoles
        ind = np.lexsort((reciprocals[:, 3], reciprocals[:, 2],
                          reciprocals[:, 1], reciprocals[:, 0]))
        reciprocals = reciprocals[ind]

        if append:
            self.configs = np.vstack((self.configs, reciprocals))
        return reciprocals

    def gen_configs_permutate(self,
                              injections_raw,
                              only_same_dipole_length=False,
                              ignore_crossed_dipoles=False,
                              silent=False):
        """ Create measurement configurations out of a pool of current
        injections.  Use only the provided dipoles for potential dipole
        selection. This means that we have always reciprocal measurements.

        Remove quadpoles where electrodes are used both as current and voltage
        dipoles.

        Parameters
        ----------
        injections_raw : Nx2 array
            current injections
        only_same_dipole_length : bool, optional
            if True, only generate permutations for the same dipole length
        ignore_crossed_dipoles : bool, optional
            If True, potential dipoles will be ignored that lie between current
            dipoles,  e.g. 1-4 3-5. In this case it is possible to not have
            full normal-reciprocal coverage.
        silent: bool, optional
            if True, do not print information on ignored configs (default:
            False)

        Returns
        -------
        configs : Nx4 array
            quadrupoles generated out of the current injections

        """
        injections = np.atleast_2d(injections_raw).astype(int)
        N = injections.shape[0]

        measurements = []

        for injection in range(0, N):
            dipole_length = np.abs(injections[injection][1] -
                                   injections[injection][0])

            # select all dipole EXCEPT for the injection dipole
            for i in set(range(0, N)) - set([injection]):
                test_dipole_length = np.abs(injections[i, :][1] -
                                            injections[i, :][0])
                if (only_same_dipole_length
                        and test_dipole_length != dipole_length):  # noqa: W503
                    continue
                quadpole = np.array(
                    [injections[injection, :], injections[i, :]]).flatten()
                if ignore_crossed_dipoles is True:
                    # check if we need to ignore this dipole
                    # Note: this could be wrong if electrode number are not
                    # ascending!
                    if (quadpole[2] > quadpole[0]
                            and quadpole[2] < quadpole[1]):  # noqa: W503
                        if not silent:
                            print('A - ignoring', quadpole)
                    elif (quadpole[3] > quadpole[0]
                          and quadpole[3] < quadpole[1]):  # noqa: W503
                        if not silent:
                            print('B - ignoring', quadpole)
                    else:
                        measurements.append(quadpole)
                else:
                    # add very quadpole
                    measurements.append(quadpole)

        # check and remove double use of electrodes
        filtered = []
        for quadpole in measurements:
            if (not set(quadpole[0:2]).isdisjoint(set(quadpole[2:4]))):
                if not silent:
                    print('Ignoring quadrupole because of ',
                          'repeated electrode use:', quadpole)
            else:
                filtered.append(quadpole)
        self.add_to_configs(filtered)
        return np.array(filtered)

    def remove_max_dipole_sep(self, maxsep=10):
        """Remove configurations with dipole separations higher than `maxsep`.

        Parameters
        ----------
        maxsep : int
            Maximum separation between both dipoles (the default is 10).
        """
        sep = np.abs(self.configs[:, 1] - self.configs[:, 2])
        self.configs = self.configs[sep <= maxsep]

    def write_configs(self, filename):
        """Write configs to file in four columns
        """
        np.savetxt(filename, self.configs, fmt='%i %i %i %i')

    def to_pg_scheme(self, container=None, positions=None):
        """Convert the configuration to a pygimli measurement scheme

        Parameters
        ----------
        container: reda.containers.ERT.ERT
            an ERT data container (we take the electrode positions from here)
        positions = None

        Returns
        -------
        data: pybert.DataContainerERT

        Examples
        --------

            import numpy as np
            from reda.configs.configManager import ConfigManager
            configs = ConfigManager(nr_of_electrodes=48)
            new_configs = configs.gen_dipole_dipole(skipc=2)
            x = np.arange(0, 48, 1)
            z = np.ones(48) * -1
            y = np.zeros(48)
            xyz = np.vstack((x, y, z)).T
            scheme = configs.to_pg_scheme(positions=xyz)
            print(scheme)


        """
        if container is None and positions is None:
            raise Exception('electrode positions are required for BERT export')

        if container is not None and container.electrodes is None:
            raise Exception('container does not contain electrode positions')

        if container is not None and positions is not None:
            raise Exception(
                'only one of container OR positions must be provided')

        if container is not None:
            elec_positions = container.electrodes.values
        elif positions is not None:
            elec_positions = positions

        opt_import("pybert", requiredFor="")
        import pybert

        # Initialize BERT DataContainer
        data = pybert.DataContainerERT()

        # Define electrodes (48 electrodes spaced by 0.5 m)
        for nr, (x, y, z) in enumerate(elec_positions):
            data.createSensor((x, y, z))

        # Define number of measurements
        data.resize(self.configs.shape[0])

        for index, token in enumerate("abmn"):
            data.set(token, self.configs[:, index].tolist())

        # account for zero indexing
        for token in "abmn":
            data.set(token, data(token) - 1)

        # np.vstack([data.get(x).array() for x in ("abmn")]).T
        return data

    def to_iris_syscal(self, filename, spacing=1):
        """Export to IRIS Instrument configuration file

        Parameters
        ----------
        filename : string
            Path to output filename
        """
        with open(filename, 'w') as fid:
            # fprintf(fod, '#\t X\t Y\t Z\n');
            fid.write('#\t X\t Y\t Z\n')
            # fprintf(fod, '%d\t %.1f\t %d\t %d\n', D');
            # loop over electrodes and assign increasing x-positions
            # TODO: use proper electrode positions, if available
            for nr in range(0, self.configs.max()):
                fid.write('{}   {}  0   0\n'.format(nr + 1, nr * spacing))

            # fprintf(fod, '#\t A\t B\t M\t N\n');
            fid.write('#\t A\t B\t M\t N\n')

            # fprintf(fod, '%d\t %d\t %d\t %d\t %d\n', C');
            for nr, config in enumerate(self.configs):
                fid.write('{}  {}  {}  {}  {}\n'.format(nr + 1, *config))

    def get_unique_current_injections(self):
        """Return all unique current injection electrode pairs

        Examples
        --------
        >>> from reda.configs.configManager import ConfigManager
        >>> cfg = ConfigManager()
        >>> cfg.nr_electrodes
        >>> cfg.nr_electrodes = 10
        >>> cfg.add_to_configs((
        ...     (1, 2, 3, 4),
        ...     (1, 2, 5, 6),
        ...     (7, 8, 9, 10)
        ... ))
        array([[ 1,  2,  3,  4],
               [ 1,  2,  5,  6],
               [ 7,  8,  9, 10]])
        >>> cfg.get_unique_current_injections()
        array([[1, 2],
               [7, 8]])

        cfg = ConfigManager()
        cfg.get_unique_current_injections()

        Returns
        -------
        unique_injections : Nx2 numpy.ndarray
            Current injections
        """
        if self.configs is None:
            return None
        return np.unique(self.configs[:, 0:2], axis=0)

    @property
    def unique_injections(self):
        return self.get_unique_current_injections()

    def remove_reciprocals(self, configs=None):
        """

        """
        if configs is None:
            configs_to_use = self.configs
        else:
            configs_to_use = configs

        indices_to_delete = []

        configs_ar = np.array(configs_to_use)

        configs_ar[:, 0:2] = np.sort(configs_ar[:, 0:2], axis=1)
        configs_ar[:, 2:4] = np.sort(configs_ar[:, 2:4], axis=1)
        for index in reversed(range(configs_ar.shape[0])):
            # print('Checking', index, configs_ar[index, :])
            # normalized reciprocal
            ar = configs_ar[index, 2]
            br = configs_ar[index, 3]
            mr = configs_ar[index, 0]
            nr = configs_ar[index, 1]

            # we only search in lower indices
            subdata = configs_ar[0:index, :]
            search = np.where(
                (
                    (subdata[:, 0] == ar) &
                    (subdata[:, 1] == br) &
                    (subdata[:, 2] == mr) &
                    (subdata[:, 3] == nr)
                )
            )[0]
            # print('search', search)
            if len(search) > 0:
                indices_to_delete.append(index)

        # print(indices_to_delete)
        for index in reversed(np.sort(indices_to_delete)):
            configs_ar = np.delete(configs_ar, index, axis=0)

        if configs is None:
            # write back
            self.configs = configs_ar

        return configs_ar



#!/usr/bin/env python
"""This is a helper class to read the primary binary measurement data of the
FZJ SIP and EIT systems SIP-04 and EIT40 (Zimmermann et al. 2008 a, b).

This is not a regular REDA-Importer as the time-domain data contained in these
binary files is not usable for geoelectric processing. However, looking at this
primary digitized data (i.e., the first digital representation of the analog
measurement signal) can help in understanding and analyzing the final SIP/sEIT
data and associated problems.

"""
import datetime
import logging
import struct
import re
import os

import matplotlib.pylab as plt
import scipy.signal
import numpy as np
import pandas as pd

# just import to set up the logger
import reda.main.logger as not_needed
not_needed


class fzj_readbin(object):
    def __init__(self, filename=None):
        """
        Parameters
        ----------
        filename : str, optional
            Filename to either the .mcf or .bin file. It is assumed that the
            corresponding .mff, .mcf, and .bin files reside in the same
            location with the same filename.

        """
        # variables to be filled when importing data
        self.filebase = None
        self.frequency_data = None
        self.frequencies = None
        self.nr_frequencies = None
        self.data = None
        self.number_injections = None
        self.injections = None

        self.logger = logging.getLogger(__name__)

        # load on initialization?
        if filename is not None and os.path.isfile(filename):
            self.import_file(filename)

    def import_file(self, filename):
        """
        Parameters
        ----------
        filename : str
            Filename to either the .mcf or .bin file. It is assumed that the
            corresponding .mff, .mcf, and .bin files reside in the same
            location with the same filename.

        """
        filebase = os.path.abspath(os.path.splitext(filename)[0])
        self.filebase = filebase

        self._read_frequencies(filebase + '.mff')

        self._read_mcf_file(filebase + '.mcf')
        self._read_data(filebase + '.bin')

    def _read_frequencies(self, mff_filename):
        testline = pd.read_csv(
            mff_filename,
            delim_whitespace=True,
            header=None,
        )
        if testline.shape[1] == 7:
            frequency_data = self._read_frequencies_sip04(mff_filename)
        else:
            frequency_data = self._read_frequencies_eit(mff_filename)

        frequency_data['fa'] = frequency_data[
            'sampling_frequency'
        ] / frequency_data['oversampling']

        frequency_data['tmax'] = frequency_data[
            'oversampling'
        ] / frequency_data[
            'sampling_frequency'
        ] * frequency_data['nr_samples']

        self.frequency_data = frequency_data
        self.frequencies = frequency_data.query(
            'inj_number == 1')['frequency'].values
        self.nr_frequencies = self.frequencies.size
        self.frequencies_unique = np.sort(
            np.unique(
                frequency_data['frequency'].values
            )
        )
        self.number_injections = int(frequency_data['inj_number'].max())

    def _read_frequencies_eit(self, mff_filename):
        frequency_data = pd.read_csv(
            mff_filename,
            delim_whitespace=True,
            header=None,
            names=[
                'delay',
                'nr_samples',
                'frequency',
                'sampling_frequency',
                'oversampling',
                'U0',
                'inj_number',
                'a',
                'b',
                'timestamp'
            ]
        )
        frequency_data['a'] = frequency_data['a'].astype(int)
        frequency_data['b'] = frequency_data['b'].astype(int)
        frequency_data['nr_samples'] = frequency_data['nr_samples'].astype(int)

        epoch = datetime.datetime(1904, 1, 1)
        frequency_data['datetime'] = [
            epoch + datetime.timedelta(
                seconds=x
            ) for x in frequency_data['timestamp'].values
        ]
        return frequency_data

    def _read_frequencies_sip04(self, mff_filename):
        frequency_data = pd.read_csv(
            mff_filename,
            delim_whitespace=True,
            header=None,
            names=[
                'delay',
                'nr_samples',
                'frequency',
                'sampling_frequency',
                'oversampling',
                'U0',
                'timestamp'
            ]
        )
        frequency_data['a'] = 1
        frequency_data['b'] = 4
        frequency_data['inj_number'] = 1
        return frequency_data

    def _read_mcf_file(self, filename):
        # encoding as iso-8859 seems to work also for utf-8
        mcf_content = open(filename, 'r', encoding='ISO-8859-1').read()

        self.NCh = int(
            re.search(
                r'NCh ([0-9]*)',
                mcf_content
            ).groups()[0]
        )

        # extract current injections
        # Note this only works with new EIT160-based mcf files
        self.injections = np.array(
            re.findall(
                r'ABMG ([0-9]?[0-9]?[0-9]) ([0-9]?[0-9]?[0-9])', mcf_content
            )
        ).astype(int)

        assert self.injections.size > 0, \
            "Error reading injections from mcf file"

    def _read_data(self, binary_file):
        data = []

        with open(binary_file, 'rb') as fid:
            for _, row in self.frequency_data.iterrows():
                N = int(row['nr_samples']) * self.NCh
                # extract 4 bytes for float16
                buffer = fid.read(4 * N)
                values = struct.unpack('>{}f'.format(N), buffer)
                subdata = np.array(values).reshape((-1, self.NCh)).T
                data.append(subdata)

        self.data = data

    def characterize(self):
        """Print a few characteristics of the loaded data"""
        if self.data is None or self.frequencies is None:
            print('No data loaded yet!')

        print('Imported from:')
        print('    {} (.bin/.mcf/.mff)'.format(self.filebase))

        # print frequency statistics
        print('Number of frequencies: {}'.format(self.nr_frequencies))
        print('Frequencies:')
        for nr, freq in enumerate(self.frequencies):
            print('{} - {} Hz'.format(nr, freq))
        print(' ')

        # print data statistics
        print('Number of channels: {}'.format(self.NCh))

        print(
            'Number of injections: {}'.format(
                self.number_injections
            )
        )

    def plot_timeseries_to_axes(
            self, axes, frequency_index, injection_number, channel,
            range_fraction=1.0, plot_style='.-',
            index_start=0,
            ):
        """
        injection_number is 1-indexed

        """
        assert len(axes) == 2

        # get the data
        index = (injection_number * self.nr_frequencies) + frequency_index
        print('index', index)
        data = self.data[index]
        fdata = self.frequency_data.iloc[index]
        # import IPython
        # IPython.embed()

        t = np.arange(0, fdata['nr_samples']) / fdata['fa']

        ax = axes[0]
        ax.grid()
        ax.set_title('Frequency: {} Hz'.format(fdata['frequency']))
        ax.set_title(
            'a-b: {}-{}, channel: {}'.format(
                int(fdata['a']),
                int(fdata['b']),
                channel,
            ), loc='right')
        print('a-b: {}-{}'.format(fdata['a'], fdata['b']))
        ax.set_ylabel('Voltage [mV]')

        # sometimes we only want to plot a fraction of the time-series, i.e.,
        # to better see higher frequencies
        index_end = min(t.size, index_start + int(t.size * range_fraction)) - 1
        print('index_end', index_end)

        x = t[index_start:index_end]
        y = data[channel, :][index_start:index_end]

        ax.plot(
            # t,
            # data[channel, :],
            x,
            y,
            plot_style,
            ms=2,
            color='k',
            linewidth=4,
        )

        ax.axhline(fdata['U0'], color='k', linestyle='dotted')
        ax.axhline(-fdata['U0'], color='k', linestyle='dotted')
        ax.set_xlabel('t [s]')

        ax = axes[1]
        ax.grid()
        # y_raw = data[channel, :]
        y_raw = y.copy()
        p = np.polyfit(x, y_raw, 1)
        trend = np.polyval(p, x)
        y = y_raw - trend
        ax.plot(x, y, '.-', color='r')

        y_transf = np.fft.rfft(y)
        y_freqs = np.fft.rfftfreq(
            n=y.size, d=1/fdata['fa']
        )

        ax.semilogx(
            y_freqs[1:],
            np.abs(y_transf[1:]) ** 2,
            '.-', ms=8, color='k'
        )
        ax.set_ylabel('$A^2$')
        print(fdata['frequency'])
        for i in range(1, 7):
            ax.axvline(
                fdata['frequency'] * i,
                color='r',
                linewidth=0.8,
                linestyle='dashed',
            )
        ax.set_xlabel('Frequency [Hz]')

    def plot_timeseries(
            self, filename, frequency_index, injection_number, channel,
            range_fraction=1.0, plot_style='.-',
            index_start=0,

            ):
        fig, axes = plt.subplots(2, 1, figsize=(16 / 2.54, 10 / 2.54))

        self.plot_timeseries_to_axes(
            axes,
            frequency_index,
            injection_number,
            channel,
            range_fraction=range_fraction,
            plot_style=plot_style,
            index_start=index_start,
        )

        fig.tight_layout()
        fig.savefig(filename, dpi=300)
        plt.close(fig)

    def plot_per_frequency(self):
        for fnr in np.arange(self.frequencies.size):
            d = self.data[fnr]
            df = pd.DataFrame(d.T)
            # dfmax = df.cummax(axis=1)
            fig, ax = plt.subplots(figsize=(20 / 2.54, 10 / 2.54))
            ax.plot(df.iloc[:, 0:40], color='gray', linewidth=2)
            ax.plot(df.iloc[:, 40:44], color='g', label='Current 40-44')
            # ax.plot(df.iloc[:, 44], color='r', label='refSignal')
            ax.legend()
            ax.set_xlabel('Sample Nr')
            ax.set_ylabel('Voltage [V]')
            ax.set_title(
                'Frequency: {} Hz'.format(self.frequencies[fnr]), loc='left')
            ax.axhline(y=9, color='k')
            ax.axhline(y=-9, color='k')
            # fig.show()
            fig.savefig('ts_f_{}.jpg'.format(fnr), dpi=300)
            plt.close(fig)

    def list_injections(self):
        """List the available injections
        """
        for index, row in enumerate(self.injections):
            print('{} - {}'.format(index + 1, row))

    def get_ts_abm(self, a, b, m, frequency):
        """Return the time series for a given set of (a, b, m electrodes)

        All values are 1-indexed!!!

        WARNING: This interface always chooses the first result related to the
        input set in case duplicate measurements are present! This relates to
        duplicate frequencies and duplicate injections.

        """
        self.logger.warn(
            'Returning time-series for: {}-{} {} at {} Hz'.format(
                a, b, m, frequency
            )
        )

        # find number of injection
        try:
            ab_nr_raw = np.where(
                (self.injections[:, 0] == a) & (self.injections[:, 1] == b)
            )[0]
            if len(ab_nr_raw) > 1:
                self.logger.warn(
                    'This injection was measured multiple times.' +
                    ' Selecting the first one.'
                )

            ab_nr = ab_nr_raw.take(0)
        except Exception:
            print('Injection not found')
            return

        index_frequency_raw = np.where(frequency == self.frequencies)[0]
        if len(index_frequency_raw) > 1:
            self.logger.warn(
                'This frequency was measured multiple times.' +
                ' Selecting the first one.'
            )
        index_frequency = index_frequency_raw.take(0)
        self.logger.info('index frequency: {}'.format(index_frequency))

        # compute starting index in data
        index_ab = self.nr_frequencies * ab_nr + index_frequency

        # add offset for m-channel
        subdata = self.data[index_ab][m - 1, :]
        return subdata

    def get_sample_times(self, frequency):
        fdata = self.frequency_data.query(
            'frequency == {}'.format(frequency)
        ).iloc[0, :]
        tmax = fdata['tmax']
        return np.linspace(0, tmax, fdata['nr_samples'].astype(int))

    def _plot_fft_analysis(
            self, measurement_index, tsdata, fft, u_peaks, noise_level,
            partnr
            ):
        """

        """
        frequency_data = self.frequency_data.iloc[measurement_index]

        tstime = self.get_sample_times(frequency_data['frequency'])
        if tstime.size > tsdata.size:
            tstime = np.split(tstime, 3)[partnr]

        fig, axes = plt.subplots(2, 1, figsize=(12 / 2.54, 9 / 2.54))
        ax = axes[0]
        ax.set_title(
            'Frequency: {} Hz'.format(frequency_data['frequency']),
            loc='left',
        )
        ax.plot(
            tstime,
            tsdata,
        )
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('Signal [V]')

        ax = axes[1]
        ax.set_title('Noise level: {}'.format(noise_level))

        fftfreq = np.fft.rfftfreq(
            tsdata.size,
            frequency_data[
                'oversampling'
            ] / frequency_data['sampling_frequency']
        )

        ax.plot(
            fftfreq[1:],
            fft[1:],
        )
        ax.scatter(
           fftfreq[u_peaks + 1],
           fft[u_peaks + 1],
           color='orange',
        )
        ax.axhline(
            y=noise_level, color='k', linestyle='dashed', label='noise level')
        ax.legend()
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel('|Amplitude|')
        ax.set_yscale('log')

        fig.tight_layout()
        return fig

    def _get_noise_level_from_fft(self, data, fs=None, **kwargs):
        # This would be a good place to try to clean-up the time-series by
        # removing the excitation frequency, harmonics, and 50/60 Hz, as
        # well as 16 2/3 train noise
        fft = np.abs(np.fft.rfft(data - data.mean()))

        if fs is not None and kwargs.get('mask_noise_harmonics', False):
            # just mask 50 Hz harmonic ranges in the fft
            freqs = np.fft.rfftfreq(data.size, 1 / fs)
            for i in range(1, 11):
                fmin = i * 50 - 5
                fmax = i * 50 + 5
                if fmin >= freqs.max() or fmax >= freqs.max():
                    continue
                index_min = np.argmin(np.abs(freqs - fmin))
                index_max = np.argmin(np.abs(freqs - fmax))
                fft[index_min:index_max] = 0

            # hack: in addition only look at data above 50 hz
            index = np.argmin(np.abs(freqs - 50))
            fft[0:index] = 0

        u_peaks, _ = scipy.signal.find_peaks(
            fft[1:], distance=kwargs.get('peak_distance', 20)
        )
        peak_values = fft[1 + u_peaks]

#         print('peak_values')
#         import IPython
#         IPython.embed()
#         exit()

        # fit a horizontal line
        noise_level = 10 ** np.polyfit(
            u_peaks, np.log10(peak_values), deg=0
        )
        return fft, u_peaks, noise_level

    def fft_analysis_one_channel(
            self, measurement_index, channel,
            split_into_three=False, plot=False, **kwargs):
        """On one specific measurement at one channel, conduct an FFT analysis
        to estimate the noise level.

        Parameters
        ----------
        measurement_index : int
            Index of injection related to index in self.frequency_data.
        channel : int
            Channel to analyze. 1-indexed.
        split_into_three : bool, optional (default: False)
            If True, apply analysis to each third of the time-series
            separately.
        plot: bool, optional (default: False)
            If True, generate plots of the time-series and noise level
        remove_excitation_frequency : bool, optional (default: False)
            .
        remove_noise_harmonics : bool (default, False)
            .

        Additional Parameters
        ---------------------
        peak_distance : int, optional (default: 20)
            Distance parameter of scipy.signal.find_peaks used to detect peaks
            in the FFT spectrum

        Returns
        -------
        noise_levels : list
            The estimated white noise leves for the parts of the time-series.
            If split_into_three is False, then the list contains only one entry
        plots : list, optional
            If generated return plots in this list
        """
        ts = self.data[measurement_index][channel - 1, :]
        if kwargs.get('remove_excitation_frequency', False):
            print('REMOVING EXCITATION SIGNAL')
            fdata = self.frequency_data.iloc[measurement_index]
            frequency = fdata['frequency']
            fs = fdata['sampling_frequency'] / fdata['oversampling']
            mage, phae = self._get_lockin(ts, frequency, fs)
            ts_signal = self._gen_signal(mage, phae, frequency, fs, ts.size)

            # ts = ts_signal
            # print(ts, ts.shape)
            ts = ts - ts_signal
            pass

        if kwargs.get('remove_noise_harmonics', False):
            print('REMOVING HARMONICS')
            pass
            # remove harmonics of signal
            fdata = self.frequency_data.iloc[measurement_index]
            frequency = fdata['frequency']
            fs = fdata['sampling_frequency'] / fdata['oversampling']
            for i in range(1, 5):
                fs_harmonic = frequency * i
                mage, phae = self._get_lockin(ts, fs_harmonic, fs)
                ts_signal = self._gen_signal(
                    mage, phae, fs_harmonic, fs, ts.size
                )
                ts = ts - ts_signal

            for i in range(1, 10):
                fs_harmonic = 50 * i
                mage, phae = self._get_lockin(ts, fs_harmonic, fs)
                ts_signal = self._gen_signal(
                    mage, phae, fs_harmonic, fs, ts.size
                )
                ts = ts - ts_signal

        if split_into_three:
            ts_parts = np.split(ts, 3)
        else:
            # analyze the full ts
            ts_parts = [ts, ]

        fdata = self.frequency_data.iloc[measurement_index]
        frequency = fdata['frequency']
        fs = fdata['sampling_frequency'] / fdata['oversampling']

        noise_levels = []
        plot_figs = []
        for partnr, part in enumerate(ts_parts):
            fft, u_peaks, noise_level = self._get_noise_level_from_fft(
                part,
                fs,
                **kwargs
            )
            noise_levels.append(noise_level)
            if plot:
                plot_figs.append(
                    self._plot_fft_analysis(
                        measurement_index,
                        part,
                        fft,
                        u_peaks,
                        noise_level,
                        partnr,
                    )
                )
        if plot:
            return noise_levels, plot_figs

        return noise_levels

    def find_swapped_measurement_indices(
            self, a, b, frequency, mean_measurement_time=None):
        """For a given set of injection electrodes and a frequency, try to find
        the two injections that will make up the final measurement (i.e., the
        regular injection (a,b) and its swapped injection (b,a).

        Parameters
        ----------
        a : int
            1. Current electrode
        b : int

            2. Current electrode
        frequency : float
            Measurement frequency
        mean_measurement_time : datetime.datetime|pandas.Timestamp
            For swapped measurements the datetime entry in the MD and EMD
            structs will be the mean time between the singular measurements.

        Returns
        -------
        findices : [int, int]
            Indices of the rows in fzj_readbin.frequency_data corresponding to
            the measurement. If only one measurement was found, then the second
            index is None.

        """

        subset = self.frequency_data.query(
            'a in ({0}, {1}) and b == ({0}, {1}) and frequency == {2}'.format(
                a, b, frequency
             )
        )
        if mean_measurement_time is None:
            self.logger.info(
                'info: mean_measurement_time not provided, will select ' +
                'earliest measurements'
            )
            mean_measurement_time = np.sort(subset['datetime'])[0]

        indices_all = np.argsort(
            np.abs(subset['datetime'] - mean_measurement_time)).values

        indices = indices_all[0:min(indices_all.size, 2)]

        # TODO: Checks
        return subset.index.values[indices]

    def plot_noise_level_for_one_injection(
            self, measurement_index, nch=None, **kwargs):
        """

        measurement_index can be found by using the search function:

            indices = self.find_swapped_measurement_indices(1, 22, 1)
            fig = plot_noise_level_for_one_injection(indices[0])

        """
        if nch is None:
            nch = self.NCh

        noise = {}
        max_values = {}
        for i in range(1, nch):
            level = self.fft_analysis_one_channel(measurement_index, i)
            noise[i] = level[0].take(0)
            ts = self.data[measurement_index][i - 1, :]
            max_value = np.max(ts - ts.mean())
            max_values[i] = max_value

        fig, ax = plt.subplots(
            figsize=kwargs.get('figsize', (12 / 2.54, 6 / 2.54))
        )
        ax.set_title(kwargs.get('title', None))
        ax.bar(noise.keys(), noise.values())
        ax.set_xlabel('Channel')
        ax.set_ylabel('Noise Level')
        ax2 = ax.twinx()
        ax2.plot(max_values.keys(), max_values.values(), '.-', color='orange')
        ax2.set_ylabel('Max. Signal [V]', color='orange')
        ax.grid()
        return fig

    @staticmethod
    def _get_lockin(data, f, fs):
        """Conduct a lockin-analysis of the given signal

        https://doi.org/10.1109/TIM.2007.908604

        Note that a phase of 0 mrad will be realized for cos functions, not
        sines!

        https://en.wikipedia.org/wiki/Phase_(waves)#Phase_shift

        Parameters
        ----------
        data : numpy.ndarray, size 3000
            Measured data
        f : float
            Analysis frequency
        fs : float
            Sampling frequency

        Returns
        -------
        magnitude : float
            Magnitude of signal at frequency f
        phase : float
            Phase of signal at frequency f [rad]

        """
        Ns = data.size
        Ni = np.arange(Ns)

        # reference signal sine
        ref_x = np.cos(2 * np.pi * f * Ni / fs)
        ref_y = np.sin(2 * np.pi * f * Ni / fs)

        # uncomment to plot reference signals
        # fig, ax = plt.subplots()
        # ax.plot(ref_x)
        # ax.plot(ref_y)
        # ax.set_xlabel('Time [s]')
        # fig.tight_layout()
        # fig.savefig('lockin_reference_signals.jpg', dpi=300)

        X = ref_x @ data / Ns
        Y = ref_y @ data / Ns

        u = np.sum(X) - 1j * np.sum(Y)
        # u = 2 * u / N
        magnitude = 2 * np.abs(u)

        # phase_mrad = np.arctan2(np.imag(u), np.real(u)) * 1000
        phase_mrad = np.arctan2(-Y, X)

        # fft
        # fft_signal = np.fft.rfft(data)
        # frequencies_fft = np.fft.rfftfreq(N, T / N)
        # print(frequencies_fft[3])

        # print('FFT entry:', fft_signal[3] * 2 / N)
        # phase_fft = (
        #     np.arctan2(
        #         np.imag(fft_signal[3]), np.real(fft_signal[3])
        #     ) + np.pi / 2
        # ) * 1000
        # print(
        #     'From fft: {}, {} mrad'.format(
        #         np.abs(fft_signal[3] * 2 / N), phase_fft))
        return magnitude, phase_mrad

    @staticmethod
    def _gen_signal(mag, pha, f, fs, Ni):
        t = np.linspace(0, Ni - 1, Ni)
        signal = mag * np.cos(2 * np.pi * f * t / fs + pha)
        return signal
